# Copyright 2020, 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import traceback
import unittest
import urllib.request
import zipfile

import aif360
import jsonschema
import numpy as np
import pandas as pd
import sklearn.metrics
import sklearn.model_selection

try:
    import cvxpy  # noqa because the import is only done as a check and flake fails

    cvxpy_installed = True
except ImportError:
    cvxpy_installed = False

try:
    import numba  # noqa because the import is only done as a check and flake fails

    numba_installed = True
except ImportError:
    numba_installed = False

try:
    import tensorflow as tf

    tensorflow_installed = True
except ImportError:
    tensorflow_installed = False

import lale.helpers
import lale.lib.aif360
import lale.lib.aif360.util
from lale.datasets.data_schemas import NDArrayWithSchema
from lale.lib.aif360 import (
    LFR,
    AdversarialDebiasing,
    CalibratedEqOddsPostprocessing,
    DisparateImpactRemover,
    EqOddsPostprocessing,
    GerryFairClassifier,
    MetaFairClassifier,
    OptimPreproc,
    PrejudiceRemover,
    Redacting,
    RejectOptionClassification,
    Reweighing,
    fair_stratified_train_test_split,
)
from lale.lib.lale import ConcatFeatures, Project
from lale.lib.sklearn import (
    FunctionTransformer,
    LinearRegression,
    LogisticRegression,
    OneHotEncoder,
)


class TestAIF360Datasets(unittest.TestCase):
    downloaded_h181 = False
    downloaded_h192 = False

    @classmethod
    def setUpClass(cls) -> None:
        cls.downloaded_h181 = cls._try_download_csv("h181.csv")
        cls.downloaded_h192 = cls._try_download_csv("h192.csv")

    @classmethod
    def tearDownClass(cls) -> None:
        if cls.downloaded_h181:
            cls._cleanup_meps("h181.csv")
        if cls.downloaded_h192:
            cls._cleanup_meps("h192.csv")

    def _attempt_dataset(
        self, X, y, fairness_info, n_rows, n_columns, set_y, di_expected
    ):
        self.assertEqual(X.shape, (n_rows, n_columns))
        self.assertEqual(y.shape, (n_rows,))
        self.assertEqual(set(y), set_y)
        di_scorer = lale.lib.aif360.disparate_impact(**fairness_info)
        di_measured = di_scorer.score_data(X=X, y_pred=y)
        self.assertAlmostEqual(di_measured, di_expected, places=3)

    def test_dataset_adult_pd_cat(self):
        X, y, fairness_info = lale.lib.aif360.fetch_adult_df(preprocess=False)
        self._attempt_dataset(X, y, fairness_info, 48_842, 14, {"<=50K", ">50K"}, 0.227)

    def test_dataset_adult_pd_num(self):
        X, y, fairness_info = lale.lib.aif360.fetch_adult_df(preprocess=True)
        self._attempt_dataset(X, y, fairness_info, 48_842, 100, {0, 1}, 0.227)

    def test_dataset_bank_pd_cat(self):
        X, y, fairness_info = lale.lib.aif360.fetch_bank_df(preprocess=False)
        self._attempt_dataset(X, y, fairness_info, 45_211, 16, {1, 2}, 0.840)

    def test_dataset_bank_pd_num(self):
        X, y, fairness_info = lale.lib.aif360.fetch_bank_df(preprocess=True)
        self._attempt_dataset(X, y, fairness_info, 45_211, 51, {0, 1}, 0.840)

    def test_dataset_compas_pd_cat(self):
        X, y, fairness_info = lale.lib.aif360.fetch_compas_df(preprocess=False)
        self._attempt_dataset(X, y, fairness_info, 6_172, 51, {0, 1}, 0.747)

    def test_dataset_compas_pd_num(self):
        X, y, fairness_info = lale.lib.aif360.fetch_compas_df(preprocess=True)
        self._attempt_dataset(X, y, fairness_info, 5_278, 10, {0, 1}, 0.687)

    def test_dataset_compas_violent_pd_cat(self):
        X, y, fairness_info = lale.lib.aif360.fetch_compas_violent_df(preprocess=False)
        self._attempt_dataset(X, y, fairness_info, 4_020, 51, {0, 1}, 0.852)

    def test_dataset_compas_violent_pd_num(self):
        X, y, fairness_info = lale.lib.aif360.fetch_compas_violent_df(preprocess=True)
        self._attempt_dataset(X, y, fairness_info, 3_377, 10, {0, 1}, 0.822)

    def test_dataset_creditg_pd_cat(self):
        X, y, fairness_info = lale.lib.aif360.fetch_creditg_df(preprocess=False)
        self._attempt_dataset(X, y, fairness_info, 1_000, 20, {"bad", "good"}, 0.748)

    def test_dataset_creditg_pd_num(self):
        X, y, fairness_info = lale.lib.aif360.fetch_creditg_df(preprocess=True)
        self._attempt_dataset(X, y, fairness_info, 1_000, 58, {0, 1}, 0.748)

    def test_dataset_nursery_pd_cat(self):
        X, y, fairness_info = lale.lib.aif360.fetch_nursery_df(preprocess=False)
        self._attempt_dataset(
            X,
            y,
            fairness_info,
            12_960,
            8,
            {"not_recom", "recommend", "very_recom", "priority", "spec_prior"},
            0.461,
        )

    def test_dataset_nursery_pd_num(self):
        X, y, fairness_info = lale.lib.aif360.fetch_nursery_df(preprocess=True)
        self._attempt_dataset(X, y, fairness_info, 12_960, 25, {0, 1}, 0.461)

    def test_dataset_ricci_pd_cat(self):
        X, y, fairness_info = lale.lib.aif360.fetch_ricci_df(preprocess=False)
        self._attempt_dataset(
            X, y, fairness_info, 118, 5, {"No promotion", "Promotion"}, 0.498
        )

    def test_dataset_ricci_pd_num(self):
        X, y, fairness_info = lale.lib.aif360.fetch_ricci_df(preprocess=True)
        self._attempt_dataset(X, y, fairness_info, 118, 6, {0, 1}, 0.498)

    def test_dataset_speeddating_pd_cat(self):
        X, y, fairness_info = lale.lib.aif360.fetch_speeddating_df(preprocess=False)
        self._attempt_dataset(X, y, fairness_info, 8_378, 122, {"0", "1"}, 0.853)

    def test_dataset_speeddating_pd_num(self):
        X, y, fairness_info = lale.lib.aif360.fetch_speeddating_df(preprocess=True)
        self._attempt_dataset(X, y, fairness_info, 8_378, 70, {0, 1}, 0.853)

    def test_dataset_boston_housing_pd_cat(self):
        X, y, fairness_info = lale.lib.aif360._fetch_boston_housing_df(preprocess=False)
        # TODO: consider better way of handling "set_y" parameter for regression problems
        self._attempt_dataset(X, y, fairness_info, 506, 13, set(y), 0.814)

    def test_dataset_boston_housing_pd_num(self):
        X, y, fairness_info = lale.lib.aif360._fetch_boston_housing_df(preprocess=True)
        # TODO: consider better way of handling "set_y" parameter for regression problems
        self._attempt_dataset(X, y, fairness_info, 506, 13, set(y), 0.814)

    def test_dataset_titanic_pd_cat(self):
        X, y, fairness_info = lale.lib.aif360.fetch_titanic_df(preprocess=False)
        self._attempt_dataset(X, y, fairness_info, 1_309, 13, {"0", "1"}, 0.263)

    def test_dataset_titanic_pd_num(self):
        X, y, fairness_info = lale.lib.aif360.fetch_titanic_df(preprocess=True)
        self._attempt_dataset(X, y, fairness_info, 1_309, 37, {0, 1}, 0.263)

    def test_dataset_tae_pd_cat(self):
        X, y, fairness_info = lale.lib.aif360.fetch_tae_df(preprocess=False)
        self._attempt_dataset(X, y, fairness_info, 151, 5, {1, 2, 3}, 0.449)

    def test_dataset_tae_pd_num(self):
        X, y, fairness_info = lale.lib.aif360.fetch_tae_df(preprocess=True)
        self._attempt_dataset(X, y, fairness_info, 151, 6, {0, 1}, 0.449)

    @classmethod
    def _try_download_csv(cls, filename):
        directory = os.path.join(
            os.path.dirname(os.path.abspath(aif360.__file__)), "data", "raw", "meps"
        )
        csv_exists = os.path.exists(
            os.path.join(
                directory,
                filename,
            )
        )
        if csv_exists:
            return False
        else:
            filename_without_extension = os.path.splitext(filename)[0]
            zip_filename = f"{filename_without_extension}ssp.zip"
            urllib.request.urlretrieve(
                f"https://meps.ahrq.gov/mepsweb/data_files/pufs/{zip_filename}",
                os.path.join(directory, zip_filename),
            )
            with zipfile.ZipFile(os.path.join(directory, zip_filename), "r") as zip_ref:
                zip_ref.extractall(directory)
            ssp_filename = f"{filename_without_extension}.ssp"
            df = pd.read_sas(os.path.join(directory, ssp_filename), format="xport")
            df.to_csv(os.path.join(directory, filename), index=False)
            return True

    @classmethod
    def _cleanup_meps(cls, filename):
        directory = os.path.join(
            os.path.dirname(os.path.abspath(aif360.__file__)), "data", "raw", "meps"
        )
        filename_without_extension = os.path.splitext(filename)[0]
        zip_filename = f"{filename_without_extension}ssp.zip"
        ssp_filename = f"{filename_without_extension}.ssp"
        os.remove(os.path.join(directory, filename))
        os.remove(os.path.join(directory, zip_filename))
        os.remove(os.path.join(directory, ssp_filename))

    def test_dataset_meps_panel19_fy2015_pd_cat(self):
        X, y, fairness_info = lale.lib.aif360.fetch_meps_panel19_fy2015_df(
            preprocess=False
        )
        self._attempt_dataset(X, y, fairness_info, 16578, 1825, {0, 1}, 0.496)

    def test_dataset_meps_panel19_fy_2015_pd_num(self):
        X, y, fairness_info = lale.lib.aif360.fetch_meps_panel19_fy2015_df(
            preprocess=True
        )
        self._attempt_dataset(X, y, fairness_info, 15830, 138, {0, 1}, 0.490)

    def test_dataset_meps_panel20_fy2015_pd_cat(self):
        X, y, fairness_info = lale.lib.aif360.fetch_meps_panel20_fy2015_df(
            preprocess=False
        )
        self._attempt_dataset(X, y, fairness_info, 18849, 1825, {0, 1}, 0.493)

    def test_dataset_meps_panel20_fy2015_pd_num(self):
        X, y, fairness_info = lale.lib.aif360.fetch_meps_panel20_fy2015_df(
            preprocess=True
        )
        self._attempt_dataset(X, y, fairness_info, 17570, 138, {0, 1}, 0.488)

    def test_dataset_meps_panel21_fy2016_pd_cat(self):
        X, y, fairness_info = lale.lib.aif360.fetch_meps_panel21_fy2016_df(
            preprocess=False
        )
        self._attempt_dataset(X, y, fairness_info, 17052, 1936, {0, 1}, 0.462)

    def test_dataset_meps_panel21_fy2016_pd_num(self):
        X, y, fairness_info = lale.lib.aif360.fetch_meps_panel21_fy2016_df(
            preprocess=True
        )
        self._attempt_dataset(X, y, fairness_info, 15675, 138, {0, 1}, 0.451)


class TestAIF360Num(unittest.TestCase):
    @classmethod
    def _creditg_pd_num(cls):
        X, y, fairness_info = lale.lib.aif360.fetch_creditg_df(preprocess=True)
        cv = lale.lib.aif360.FairStratifiedKFold(**fairness_info, n_splits=3)
        splits = []
        lr = LogisticRegression()
        for train, test in cv.split(X, y):
            train_X, train_y = lale.helpers.split_with_schemas(lr, X, y, train)
            assert isinstance(train_X, pd.DataFrame), type(train_X)
            assert isinstance(train_y, pd.Series), type(train_y)
            test_X, test_y = lale.helpers.split_with_schemas(lr, X, y, test, train)
            assert isinstance(test_X, pd.DataFrame), type(test_X)
            assert isinstance(test_y, pd.Series), type(test_y)
            splits.append(
                {
                    "train_X": train_X,
                    "train_y": train_y,
                    "test_X": test_X,
                    "test_y": test_y,
                }
            )
        result = {"splits": splits, "fairness_info": fairness_info}
        return result

    @classmethod
    def _creditg_np_num(cls):
        train_X = cls.creditg_pd_num["splits"][0]["train_X"].to_numpy()
        train_y = cls.creditg_pd_num["splits"][0]["train_y"].to_numpy()
        test_X = cls.creditg_pd_num["splits"][0]["test_X"].to_numpy()
        test_y = cls.creditg_pd_num["splits"][0]["test_y"].to_numpy()
        assert isinstance(train_X, np.ndarray), type(train_X)
        assert not isinstance(train_X, NDArrayWithSchema), type(train_X)
        assert isinstance(train_y, np.ndarray), type(train_y)
        assert not isinstance(train_y, NDArrayWithSchema), type(train_y)
        assert isinstance(test_X, np.ndarray), type(test_X)
        assert not isinstance(test_X, NDArrayWithSchema), type(test_X)
        assert isinstance(test_y, np.ndarray), type(test_y)
        assert not isinstance(test_y, NDArrayWithSchema), type(test_y)
        pd_columns = cls.creditg_pd_num["splits"][0]["train_X"].columns
        fairness_info = {
            "favorable_labels": [1],
            "protected_attributes": [
                {
                    "feature": pd_columns.get_loc("sex"),
                    "reference_group": [1],
                },
                {
                    "feature": pd_columns.get_loc("age"),
                    "reference_group": [1],
                },
            ],
        }
        result = {
            "train_X": train_X,
            "train_y": train_y,
            "test_X": test_X,
            "test_y": test_y,
            "fairness_info": fairness_info,
        }
        return result

    @classmethod
    def _boston_pd_num(cls):
        # TODO: Consider investigating test failure when preprocess is set to True
        # (eo_diff is not less than 0 in this case; perhaps regression model learns differently?)
        orig_X, orig_y, fairness_info = lale.lib.aif360._fetch_boston_housing_df(
            preprocess=False
        )
        train_X, test_X, train_y, test_y = sklearn.model_selection.train_test_split(
            orig_X, orig_y, test_size=0.33, random_state=42
        )
        assert isinstance(train_X, pd.DataFrame), type(train_X)
        assert isinstance(train_y, pd.Series), type(train_y)
        assert isinstance(test_X, pd.DataFrame), type(test_X)
        assert isinstance(test_y, pd.Series), type(test_y)
        result = {
            "train_X": train_X,
            "train_y": train_y,
            "test_X": test_X,
            "test_y": test_y,
            "fairness_info": fairness_info,
        }
        return result

    @classmethod
    def _boston_np_num(cls):
        train_X = cls.boston_pd_num["train_X"].to_numpy()
        train_y = cls.boston_pd_num["train_y"].to_numpy()
        test_X = cls.boston_pd_num["test_X"].to_numpy()
        test_y = cls.boston_pd_num["test_y"].to_numpy()
        assert isinstance(train_X, np.ndarray), type(train_X)
        assert not isinstance(train_X, NDArrayWithSchema), type(train_X)
        assert isinstance(train_y, np.ndarray), type(train_y)
        assert not isinstance(train_y, NDArrayWithSchema), type(train_y)
        assert isinstance(test_X, np.ndarray), type(test_X)
        assert not isinstance(test_X, NDArrayWithSchema), type(test_X)
        assert isinstance(test_y, np.ndarray), type(test_y)
        assert not isinstance(test_y, NDArrayWithSchema), type(test_y)
        pd_columns = cls.boston_pd_num["train_X"].columns
        # pulling attributes off of stored fairness_info to avoid recomputing medians
        fairness_info = {
            "favorable_labels": cls.boston_pd_num["fairness_info"]["favorable_labels"],
            "protected_attributes": [
                {
                    "feature": pd_columns.get_loc("B"),
                    "reference_group": cls.boston_pd_num["fairness_info"][
                        "protected_attributes"
                    ][0]["reference_group"],
                },
            ],
        }
        result = {
            "train_X": train_X,
            "train_y": train_y,
            "test_X": test_X,
            "test_y": test_y,
            "fairness_info": fairness_info,
        }
        return result

    @classmethod
    def setUpClass(cls):
        cls.creditg_pd_num = cls._creditg_pd_num()
        cls.creditg_np_num = cls._creditg_np_num()
        cls.boston_pd_num = cls._boston_pd_num()
        cls.boston_np_num = cls._boston_np_num()

    def test_fair_stratified_train_test_split(self):
        X = self.creditg_np_num["train_X"]
        y = self.creditg_np_num["train_y"]
        fairness_info = self.creditg_np_num["fairness_info"]
        z = range(X.shape[0])
        (
            train_X,
            test_X,
            train_y,
            test_y,
            train_z,
            test_z,
        ) = fair_stratified_train_test_split(X, y, z, **fairness_info)
        self.assertEqual(train_X.shape[0], train_y.shape[0])
        self.assertEqual(train_X.shape[0], len(train_z))
        self.assertEqual(test_X.shape[0], test_y.shape[0])
        self.assertEqual(test_X.shape[0], len(test_z))
        self.assertEqual(train_X.shape[0] + test_X.shape[0], X.shape[0])

    def _attempt_scorers(self, fairness_info, estimator, test_X, test_y):
        fi = fairness_info
        disparate_impact_scorer = lale.lib.aif360.disparate_impact(**fi)
        impact = disparate_impact_scorer(estimator, test_X, test_y)
        self.assertLess(impact, 0.9)
        if estimator.is_classifier():
            combined_scorer = lale.lib.aif360.accuracy_and_disparate_impact(**fi)
            combined = combined_scorer(estimator, test_X, test_y)
            self.assertLess(0.0, combined)
            self.assertLess(combined, 1.0)
        else:
            combined_scorer = lale.lib.aif360.r2_and_disparate_impact(**fi)
            combined = combined_scorer(estimator, test_X, test_y)
            self.assertLess(0.0, combined)
            self.assertLess(combined, 1.0)
        parity_scorer = lale.lib.aif360.statistical_parity_difference(**fi)
        parity = parity_scorer(estimator, test_X, test_y)
        self.assertLess(parity, 0.0)
        eo_diff_scorer = lale.lib.aif360.equal_opportunity_difference(**fi)
        eo_diff = eo_diff_scorer(estimator, test_X, test_y)
        self.assertLess(eo_diff, 0.0)
        ao_diff_scorer = lale.lib.aif360.average_odds_difference(**fi)
        ao_diff = ao_diff_scorer(estimator, test_X, test_y)
        self.assertLess(ao_diff, 0.1)
        theil_index_scorer = lale.lib.aif360.theil_index(**fi)
        theil_index = theil_index_scorer(estimator, test_X, test_y)
        self.assertGreater(theil_index, 0.1)
        symm_di_scorer = lale.lib.aif360.symmetric_disparate_impact(**fi)
        symm_di = symm_di_scorer(estimator, test_X, test_y)
        self.assertLess(symm_di, 0.9)

    def test_scorers_pd_num(self):
        fairness_info = self.creditg_pd_num["fairness_info"]
        trainable = LogisticRegression(max_iter=1000)
        train_X = self.creditg_pd_num["splits"][0]["train_X"]
        train_y = self.creditg_pd_num["splits"][0]["train_y"]
        trained = trainable.fit(train_X, train_y)
        test_X = self.creditg_pd_num["splits"][0]["test_X"]
        test_y = self.creditg_pd_num["splits"][0]["test_y"]
        self._attempt_scorers(fairness_info, trained, test_X, test_y)

    def test_scorers_np_num(self):
        fairness_info = self.creditg_np_num["fairness_info"]
        trainable = LogisticRegression(max_iter=1000)
        train_X = self.creditg_np_num["train_X"]
        train_y = self.creditg_np_num["train_y"]
        trained = trainable.fit(train_X, train_y)
        test_X = self.creditg_np_num["test_X"]
        test_y = self.creditg_np_num["test_y"]
        self._attempt_scorers(fairness_info, trained, test_X, test_y)

    def test_scorers_regression_pd_num(self):
        fairness_info = self.boston_pd_num["fairness_info"]
        trainable = LinearRegression()
        train_X = self.boston_pd_num["train_X"]
        train_y = self.boston_pd_num["train_y"]
        trained = trainable.fit(train_X, train_y)
        test_X = self.boston_pd_num["test_X"]
        test_y = self.boston_pd_num["test_y"]
        self._attempt_scorers(fairness_info, trained, test_X, test_y)

    def test_scorers_regression_np_num(self):
        fairness_info = self.boston_np_num["fairness_info"]
        trainable = LinearRegression()
        train_X = self.boston_np_num["train_X"]
        train_y = self.boston_np_num["train_y"]
        trained = trainable.fit(train_X, train_y)
        test_X = self.boston_np_num["test_X"]
        test_y = self.boston_np_num["test_y"]
        self._attempt_scorers(fairness_info, trained, test_X, test_y)

    def test_scorers_combine_acc(self):
        dummy_fairness_info = {
            "favorable_labels": ["fav"],
            "protected_attributes": [{"feature": "prot", "reference_group": ["ref"]}],
        }
        scorer = lale.lib.aif360.accuracy_and_disparate_impact(**dummy_fairness_info)
        for acc in [0.2, 0.8, 1]:
            for di in [0.7, 0.9, 1.0]:
                score = scorer._combine(acc, di)
                self.assertLess(0.0, score)
                self.assertLessEqual(score, 1.0)
            for di in [0.0, float("inf"), float("-inf"), float("nan")]:
                score = scorer._combine(acc, di)
                self.assertEqual(score, 0.5 * acc)

    def test_scorers_combine_r2(self):
        dummy_fairness_info = {
            "favorable_labels": ["fav"],
            "protected_attributes": [{"feature": "prot", "reference_group": ["ref"]}],
        }
        scorer = lale.lib.aif360.r2_and_disparate_impact(**dummy_fairness_info)
        for r2 in [-2, 0, 0.5, 1]:
            for di in [0.7, 0.9, 1.0]:
                score = scorer._combine(r2, di)
                self.assertLess(0.0, score)
                self.assertLessEqual(score, 1.0)
            for di in [0.0, float("inf"), float("-inf"), float("nan")]:
                score = scorer._combine(r2, di)
                self.assertEqual(score, 0.5 / (2.0 - r2))

    def _attempt_remi_creditg_pd_num(
        self, fairness_info, trainable_remi, min_di, max_di
    ):
        splits = self.creditg_pd_num["splits"]
        disparate_impact_scorer = lale.lib.aif360.disparate_impact(**fairness_info)
        di_list = []
        for split in splits:
            if tensorflow_installed:  # for AdversarialDebiasing
                tf.compat.v1.reset_default_graph()
                tf.compat.v1.disable_eager_execution()
            train_X = split["train_X"]
            train_y = split["train_y"]
            trained_remi = trainable_remi.fit(train_X, train_y)
            test_X = split["test_X"]
            test_y = split["test_y"]
            di_list.append(disparate_impact_scorer(trained_remi, test_X, test_y))
        di = pd.Series(di_list)
        _, _, function_name, _ = traceback.extract_stack()[-2]
        print(f"disparate impact {di.mean():.3f} +- {di.std():.3f} {function_name}")
        if min_di > 0:
            self.assertLessEqual(min_di, di.mean())
            self.assertLessEqual(di.mean(), max_di)

    def test_disparate_impact_remover_np_num(self):
        fairness_info = self.creditg_np_num["fairness_info"]
        trainable_orig = LogisticRegression(max_iter=1000)
        trainable_remi = DisparateImpactRemover(**fairness_info) >> trainable_orig
        train_X = self.creditg_np_num["train_X"]
        train_y = self.creditg_np_num["train_y"]
        trained_orig = trainable_orig.fit(train_X, train_y)
        trained_remi = trainable_remi.fit(train_X, train_y)
        test_X = self.creditg_np_num["test_X"]
        test_y = self.creditg_np_num["test_y"]
        disparate_impact_scorer = lale.lib.aif360.disparate_impact(**fairness_info)
        impact_orig = disparate_impact_scorer(trained_orig, test_X, test_y)
        self.assertTrue(0.6 < impact_orig < 1.0, f"impact_orig {impact_orig}")
        impact_remi = disparate_impact_scorer(trained_remi, test_X, test_y)
        self.assertTrue(0.8 < impact_remi < 1.0, f"impact_remi {impact_remi}")

    def test_adversarial_debiasing_pd_num(self):
        if tensorflow_installed:
            fairness_info = self.creditg_pd_num["fairness_info"]
            tf.compat.v1.reset_default_graph()
            trainable_remi = AdversarialDebiasing(**fairness_info)
            self._attempt_remi_creditg_pd_num(fairness_info, trainable_remi, 0.0, 1.5)

    def test_calibrated_eq_odds_postprocessing_pd_num(self):
        fairness_info = self.creditg_pd_num["fairness_info"]
        estim = LogisticRegression(max_iter=1000)
        trainable_remi = CalibratedEqOddsPostprocessing(
            **fairness_info, estimator=estim
        )
        self._attempt_remi_creditg_pd_num(fairness_info, trainable_remi, 0.65, 0.85)

    def test_disparate_impact_remover_pd_num(self):
        fairness_info = self.creditg_pd_num["fairness_info"]
        trainable_remi = DisparateImpactRemover(**fairness_info) >> LogisticRegression(
            max_iter=1000
        )
        self._attempt_remi_creditg_pd_num(fairness_info, trainable_remi, 0.78, 0.88)

    def test_eq_odds_postprocessing_pd_num(self):
        fairness_info = self.creditg_pd_num["fairness_info"]
        estim = LogisticRegression(max_iter=1000)
        trainable_remi = EqOddsPostprocessing(**fairness_info, estimator=estim)
        self._attempt_remi_creditg_pd_num(fairness_info, trainable_remi, 0.82, 1.02)

    def test_gerry_fair_classifier_pd_num(self):
        fairness_info = self.creditg_pd_num["fairness_info"]
        trainable_remi = GerryFairClassifier(**fairness_info)
        self._attempt_remi_creditg_pd_num(fairness_info, trainable_remi, 0.677, 0.678)

    def test_lfr_pd_num(self):
        if numba_installed:
            fairness_info = self.creditg_pd_num["fairness_info"]
            trainable_remi = LFR(**fairness_info) >> LogisticRegression(max_iter=1000)
            self._attempt_remi_creditg_pd_num(fairness_info, trainable_remi, 0.95, 1.05)

    def test_meta_fair_classifier_pd_num(self):
        if aif360.__version__ >= "4.0.0":
            fairness_info = self.creditg_pd_num["fairness_info"]
            trainable_remi = MetaFairClassifier(**fairness_info)
            self._attempt_remi_creditg_pd_num(fairness_info, trainable_remi, 0.62, 0.87)

    def test_prejudice_remover_pd_num(self):
        fairness_info = self.creditg_pd_num["fairness_info"]
        trainable_remi = PrejudiceRemover(**fairness_info)
        self._attempt_remi_creditg_pd_num(fairness_info, trainable_remi, 0.73, 0.83)

    def test_redacting_pd_num(self):
        fairness_info = self.creditg_pd_num["fairness_info"]
        redacting = Redacting(**fairness_info)
        logistic_regression = LogisticRegression(max_iter=1000)
        trainable_remi = redacting >> logistic_regression
        self._attempt_remi_creditg_pd_num(fairness_info, trainable_remi, 0.78, 0.94)

    def test_reject_option_classification_pd_num(self):
        fairness_info = self.creditg_pd_num["fairness_info"]
        estim = LogisticRegression(max_iter=1000)
        trainable_remi = RejectOptionClassification(**fairness_info, estimator=estim)
        self._attempt_remi_creditg_pd_num(fairness_info, trainable_remi, 0.88, 0.98)

    def test_reweighing_pd_num(self):
        fairness_info = self.creditg_pd_num["fairness_info"]
        estim = LogisticRegression(max_iter=1000)
        trainable_remi = Reweighing(estimator=estim, **fairness_info)
        self._attempt_remi_creditg_pd_num(fairness_info, trainable_remi, 0.82, 0.92)

    def test_sans_mitigation_pd_num(self):
        fairness_info = self.creditg_pd_num["fairness_info"]
        trainable_remi = LogisticRegression(max_iter=1000)
        self._attempt_remi_creditg_pd_num(fairness_info, trainable_remi, 0.5, 1.0)


class TestAIF360Cat(unittest.TestCase):
    @classmethod
    def _prep_pd_cat(cls):
        result = (
            (
                Project(columns={"type": "string"})
                >> OneHotEncoder(handle_unknown="ignore")
            )
            & Project(columns={"type": "number"})
        ) >> ConcatFeatures
        return result

    @classmethod
    def _creditg_pd_cat(cls):
        X, y, fairness_info = lale.lib.aif360.fetch_creditg_df(preprocess=False)
        cv = lale.lib.aif360.FairStratifiedKFold(**fairness_info, n_splits=3)
        splits = []
        lr = LogisticRegression()
        for train, test in cv.split(X, y):
            train_X, train_y = lale.helpers.split_with_schemas(lr, X, y, train)
            assert isinstance(train_X, pd.DataFrame), type(train_X)
            assert isinstance(train_y, pd.Series), type(train_y)
            test_X, test_y = lale.helpers.split_with_schemas(lr, X, y, test, train)
            assert isinstance(test_X, pd.DataFrame), type(test_X)
            assert isinstance(test_y, pd.Series), type(test_y)
            splits.append(
                {
                    "train_X": train_X,
                    "train_y": train_y,
                    "test_X": test_X,
                    "test_y": test_y,
                }
            )
        result = {"splits": splits, "fairness_info": fairness_info}
        return result

    @classmethod
    def _creditg_np_cat(cls):
        train_X = cls.creditg_pd_cat["splits"][0]["train_X"].to_numpy()
        train_y = cls.creditg_pd_cat["splits"][0]["train_y"].to_numpy()
        test_X = cls.creditg_pd_cat["splits"][0]["test_X"].to_numpy()
        test_y = cls.creditg_pd_cat["splits"][0]["test_y"].to_numpy()
        assert isinstance(train_X, np.ndarray), type(train_X)
        assert not isinstance(train_X, NDArrayWithSchema), type(train_X)
        assert isinstance(train_y, np.ndarray), type(train_y)
        assert not isinstance(train_y, NDArrayWithSchema), type(train_y)
        pd_columns = cls.creditg_pd_cat["splits"][0]["train_X"].columns
        pd_fav_labels = cls.creditg_pd_cat["fairness_info"]["favorable_labels"]
        pd_prot_attrs = cls.creditg_pd_cat["fairness_info"]["protected_attributes"]
        fairness_info = {
            "favorable_labels": pd_fav_labels,
            "protected_attributes": [
                {
                    "feature": pd_columns.get_loc("personal_status"),
                    "reference_group": pd_prot_attrs[0]["reference_group"],
                },
                {
                    "feature": pd_columns.get_loc("age"),
                    "reference_group": pd_prot_attrs[1]["reference_group"],
                },
            ],
        }
        result = {
            "train_X": train_X,
            "train_y": train_y,
            "test_X": test_X,
            "test_y": test_y,
            "fairness_info": fairness_info,
        }
        return result

    @classmethod
    def _creditg_pd_ternary(cls):
        X, y, _ = lale.lib.aif360.fetch_creditg_df(preprocess=False)
        fairness_info = {
            "favorable_labels": ["good"],
            "protected_attributes": [
                {
                    "feature": "personal_status",
                    "reference_group": ["male div/sep", "male mar/wid", "male single"],
                    "monitored_group": ["female div/dep/mar"],
                },
                {
                    "feature": "age",
                    "reference_group": [[26, 1000]],
                    "monitored_group": [[1, 23]],
                },
            ],
            "unfavorable_labels": ["bad"],
        }
        cv = lale.lib.aif360.FairStratifiedKFold(**fairness_info, n_splits=3)
        splits = []
        lr = LogisticRegression()
        for train, test in cv.split(X, y):
            train_X, train_y = lale.helpers.split_with_schemas(lr, X, y, train)
            assert isinstance(train_X, pd.DataFrame), type(train_X)
            assert isinstance(train_y, pd.Series), type(train_y)
            test_X, test_y = lale.helpers.split_with_schemas(lr, X, y, test, train)
            assert isinstance(test_X, pd.DataFrame), type(test_X)
            assert isinstance(test_y, pd.Series), type(test_y)
            splits.append(
                {
                    "train_X": train_X,
                    "train_y": train_y,
                    "test_X": test_X,
                    "test_y": test_y,
                }
            )
        result = {"splits": splits, "fairness_info": fairness_info}
        return result

    @classmethod
    def setUpClass(cls):
        cls.prep_pd_cat = cls._prep_pd_cat()
        cls.creditg_pd_cat = cls._creditg_pd_cat()
        cls.creditg_np_cat = cls._creditg_np_cat()
        cls.creditg_pd_ternary = cls._creditg_pd_ternary()

    def test_encoder_pd_cat(self):
        info = self.creditg_pd_cat["fairness_info"]
        orig_X = self.creditg_pd_cat["splits"][0]["train_X"]
        encoder_separate = lale.lib.aif360.ProtectedAttributesEncoder(
            protected_attributes=info["protected_attributes"]
        )
        csep_X = encoder_separate.transform(orig_X)
        encoder_and = lale.lib.aif360.ProtectedAttributesEncoder(
            protected_attributes=info["protected_attributes"], combine="and"
        )
        cand_X = encoder_and.transform(orig_X)
        for i in orig_X.index:
            orig_row = orig_X.loc[i]
            csep_row = csep_X.loc[i]
            cand_row = cand_X.loc[i]
            cand_name = list(cand_X.columns)[0]
            self.assertEqual(
                1 if orig_row["personal_status"].startswith("male") else 0,
                csep_row["personal_status"],
            )
            self.assertEqual(1 if orig_row["age"] >= 26 else 0, csep_row["age"])
            self.assertEqual(
                cand_row[cand_name],
                1 if csep_row["personal_status"] == 1 == csep_row["age"] else 0,
            )

    def test_encoder_np_cat(self):
        info = self.creditg_np_cat["fairness_info"]
        orig_X = self.creditg_np_cat["train_X"]
        encoder = lale.lib.aif360.ProtectedAttributesEncoder(
            protected_attributes=info["protected_attributes"]
        )
        conv_X = encoder.transform(orig_X)
        for i in range(orig_X.shape[0]):
            self.assertEqual(
                1 if orig_X[i, 8].startswith("male") else 0,
                conv_X.at[i, "f8"],
            )
            self.assertEqual(1 if orig_X[i, 12] >= 26 else 0, conv_X.at[i, "f12"])

    def test_encoder_pd_ternary(self):
        info = self.creditg_pd_ternary["fairness_info"]
        orig_X = self.creditg_pd_ternary["splits"][0]["train_X"]
        encoder_separate = lale.lib.aif360.ProtectedAttributesEncoder(
            protected_attributes=info["protected_attributes"]
        )
        csep_X = encoder_separate.transform(orig_X)
        encoder_and = lale.lib.aif360.ProtectedAttributesEncoder(
            protected_attributes=info["protected_attributes"], combine="and"
        )
        cand_X = encoder_and.transform(orig_X)
        for i in orig_X.index:
            orig_row = orig_X.loc[i]
            csep_row = csep_X.loc[i]
            cand_row = cand_X.loc[i]
            cand_name = list(cand_X.columns)[0]
            self.assertEqual(
                1
                if orig_row["personal_status"].startswith("male")
                else 0
                if orig_row["personal_status"].startswith("female")
                else 0.5,
                csep_row["personal_status"],
                f"personal_status {orig_row['personal_status']}",
            )
            self.assertEqual(
                1
                if 26 <= orig_row["age"] <= 1000
                else 0
                if 1 <= orig_row["age"] <= 23
                else 0.5,
                csep_row["age"],
                f"age {orig_row['age']}",
            )
            self.assertEqual(
                cand_row[cand_name],
                min(csep_row["personal_status"], csep_row["age"]),
                f"age {orig_row['age']}, personal_status {orig_row['personal_status']}",
            )

    def test_column_for_stratification(self):
        fairness_info = self.creditg_pd_cat["fairness_info"]
        train_X = self.creditg_pd_cat["splits"][0]["train_X"]
        train_y = self.creditg_pd_cat["splits"][0]["train_y"]
        stratify = lale.lib.aif360.util._column_for_stratification(
            train_X, train_y, **fairness_info, unfavorable_labels=None
        )
        for i in train_X.index:
            male = train_X.loc[i]["personal_status"].startswith("male")
            old = train_X.loc[i]["age"] >= 26
            favorable = train_y.loc[i] == "good"
            strat = stratify.loc[i]
            self.assertEqual(male, strat[0] == "T")
            self.assertEqual(old, strat[1] == "T")
            self.assertEqual(favorable, strat[2] == "T")

    def test_fair_stratified_train_test_split(self):
        X, y, fairness_info = lale.lib.aif360.fetch_creditg_df(preprocess=False)
        z = range(X.shape[0])
        (
            train_X,
            test_X,
            train_y,
            test_y,
            train_z,
            test_z,
        ) = fair_stratified_train_test_split(X, y, z, **fairness_info)
        self.assertEqual(train_X.shape[0], train_y.shape[0])
        self.assertEqual(train_X.shape[0], len(train_z))
        self.assertEqual(test_X.shape[0], test_y.shape[0])
        self.assertEqual(test_X.shape[0], len(test_z))

    def _attempt_scorers(self, fairness_info, estimator, test_X, test_y):
        fi = fairness_info
        disparate_impact_scorer = lale.lib.aif360.disparate_impact(**fi)
        impact = disparate_impact_scorer(estimator, test_X, test_y)
        self.assertLess(impact, 0.9)
        if estimator.is_classifier():
            combined_scorer = lale.lib.aif360.accuracy_and_disparate_impact(**fi)
            combined = combined_scorer(estimator, test_X, test_y)
            self.assertLess(0.0, combined)
            self.assertLess(combined, 1.0)
        else:
            combined_scorer = lale.lib.aif360.r2_and_disparate_impact(**fi)
            combined = combined_scorer(estimator, test_X, test_y)
            self.assertLess(0.0, combined)
            self.assertLess(combined, 1.0)
        parity_scorer = lale.lib.aif360.statistical_parity_difference(**fi)
        parity = parity_scorer(estimator, test_X, test_y)
        self.assertLess(parity, 0.0)
        eo_diff_scorer = lale.lib.aif360.equal_opportunity_difference(**fi)
        eo_diff = eo_diff_scorer(estimator, test_X, test_y)
        self.assertLess(eo_diff, 0.0)
        ao_diff_scorer = lale.lib.aif360.average_odds_difference(**fi)
        ao_diff = ao_diff_scorer(estimator, test_X, test_y)
        self.assertLess(ao_diff, 0.1)
        theil_index_scorer = lale.lib.aif360.theil_index(**fi)
        theil_index = theil_index_scorer(estimator, test_X, test_y)
        self.assertGreater(theil_index, 0.1)
        symm_di_scorer = lale.lib.aif360.symmetric_disparate_impact(**fi)
        symm_di = symm_di_scorer(estimator, test_X, test_y)
        self.assertLess(symm_di, 0.9)

    def test_scorers_pd_cat(self):
        fairness_info = self.creditg_pd_cat["fairness_info"]
        trainable = self.prep_pd_cat >> LogisticRegression(max_iter=1000)
        train_X = self.creditg_pd_cat["splits"][0]["train_X"]
        train_y = self.creditg_pd_cat["splits"][0]["train_y"]
        trained = trainable.fit(train_X, train_y)
        test_X = self.creditg_pd_cat["splits"][0]["test_X"]
        test_y = self.creditg_pd_cat["splits"][0]["test_y"]
        self._attempt_scorers(fairness_info, trained, test_X, test_y)

    def test_scorers_np_cat(self):
        fairness_info = self.creditg_np_cat["fairness_info"]
        train_X = self.creditg_np_cat["train_X"]
        train_y = self.creditg_np_cat["train_y"]
        cat_columns, num_columns = [], []
        for i in range(train_X.shape[1]):
            try:
                _ = train_X[:, i].astype(np.float64)
                num_columns.append(i)
            except ValueError:
                cat_columns.append(i)
        trainable = (
            (
                (Project(columns=cat_columns) >> OneHotEncoder(handle_unknown="ignore"))
                & (
                    Project(columns=num_columns)
                    >> FunctionTransformer(func=lambda x: x.astype(np.float64))
                )
            )
            >> ConcatFeatures
            >> LogisticRegression(max_iter=1000)
        )
        trained = trainable.fit(train_X, train_y)
        test_X = self.creditg_np_cat["test_X"]
        test_y = self.creditg_np_cat["test_y"]
        self._attempt_scorers(fairness_info, trained, test_X, test_y)

    def test_scorers_pd_ternary(self):
        fairness_info = self.creditg_pd_ternary["fairness_info"]
        trainable = self.prep_pd_cat >> LogisticRegression(max_iter=1000)
        train_X = self.creditg_pd_ternary["splits"][0]["train_X"]
        train_y = self.creditg_pd_ternary["splits"][0]["train_y"]
        trained = trainable.fit(train_X, train_y)
        test_X = self.creditg_pd_ternary["splits"][0]["test_X"]
        test_y = self.creditg_pd_ternary["splits"][0]["test_y"]
        self._attempt_scorers(fairness_info, trained, test_X, test_y)

    def test_scorers_warn(self):
        fairness_info = {
            "favorable_labels": ["good"],
            "protected_attributes": [{"feature": "age", "reference_group": [1]}],
        }
        trainable = self.prep_pd_cat >> LogisticRegression(max_iter=1000)
        train_X = self.creditg_pd_cat["splits"][0]["train_X"]
        train_y = self.creditg_pd_cat["splits"][0]["train_y"]
        trained = trainable.fit(train_X, train_y)
        test_X = self.creditg_pd_cat["splits"][0]["test_X"]
        test_y = self.creditg_pd_cat["splits"][0]["test_y"]
        disparate_impact_scorer = lale.lib.aif360.disparate_impact(**fairness_info)
        with self.assertLogs(lale.lib.aif360.util.logger) as log_context_manager:
            impact = disparate_impact_scorer(trained, test_X, test_y)
        self.assertRegex(log_context_manager.output[-1], "is ill-defined")
        self.assertTrue(np.isnan(impact))

    def test_scorers_warn2(self):
        fairness_info = {
            "favorable_labels": ["good"],
            "unfavorable_labels": ["good"],
            "protected_attributes": [
                {"feature": "age", "reference_group": [[26, 1000]]}
            ],
        }
        with self.assertLogs(lale.lib.aif360.util.logger) as log_context_manager:
            _ = lale.lib.aif360.disparate_impact(**fairness_info)
        self.assertRegex(
            log_context_manager.output[-1],
            "overlap between favorable labels and unfavorable labels on 'good' and 'good'",
        )

    def test_scorers_warn3(self):
        fairness_info = {
            "favorable_labels": ["good"],
            "protected_attributes": [
                {"feature": "age", "reference_group": [[1, 2, 3]]}
            ],
        }
        with self.assertRaises(jsonschema.ValidationError):
            _ = lale.lib.aif360.disparate_impact(**fairness_info)

    def test_scorers_warn4(self):
        fairness_info = {
            "favorable_labels": ["good"],
            "protected_attributes": [
                {
                    "feature": "age",
                    "reference_group": [[20, 40]],
                    "monitored_group": [30],
                }
            ],
        }
        with self.assertLogs(lale.lib.aif360.util.logger) as log_context_manager:
            _ = lale.lib.aif360.disparate_impact(**fairness_info)
        self.assertRegex(
            log_context_manager.output[-1],
            "overlap between reference group and monitored group of feature 'age'",
        )

    def test_scorers_ternary_nonexhaustive(self):
        X, y, fairness_info = lale.lib.aif360.fetch_nursery_df(preprocess=False)
        self.assertEqual(
            set(y), {"not_recom", "recommend", "very_recom", "priority", "spec_prior"}
        )
        fairness_info = {**fairness_info, "unfavorable_labels": ["not_recom"]}
        self.assertEqual(
            set(
                fairness_info["favorable_labels"] + fairness_info["unfavorable_labels"]
            ),
            {"spec_prior", "not_recom"},
        )
        di_scorer = lale.lib.aif360.disparate_impact(**fairness_info)
        di_measured = di_scorer.score_data(X=X, y_pred=y)
        self.assertAlmostEqual(di_measured, 0.461, places=3)
        sp_scorer = lale.lib.aif360.statistical_parity_difference(**fairness_info)
        sp_measured = sp_scorer.score_data(X=X, y_pred=y)
        self.assertAlmostEqual(sp_measured, -0.205, places=3)
        sdi_scorer = lale.lib.aif360.symmetric_disparate_impact(**fairness_info)
        sdi_measured = sdi_scorer.score_data(X=X, y_pred=y)
        self.assertAlmostEqual(sdi_measured, 0.461, places=3)
        adi_scorer = lale.lib.aif360.accuracy_and_disparate_impact(**fairness_info)
        adi_measured = adi_scorer.score_data(X=X, y_pred=y, y_true=y)
        self.assertAlmostEqual(adi_measured, 0.731, places=3)
        ao_scorer = lale.lib.aif360.average_odds_difference(**fairness_info)
        with self.assertRaisesRegex(ValueError, "unexpected labels"):
            _ = ao_scorer.score_data(X=X, y_pred=y)

    def _attempt_remi_creditg_pd_cat(
        self, fairness_info, trainable_remi, min_di, max_di
    ):
        splits = self.creditg_pd_cat["splits"]
        disparate_impact_scorer = lale.lib.aif360.disparate_impact(**fairness_info)
        di_list = []
        for split in splits:
            if tensorflow_installed:  # for AdversarialDebiasing
                tf.compat.v1.reset_default_graph()
            train_X = split["train_X"]
            train_y = split["train_y"]
            trained_remi = trainable_remi.fit(train_X, train_y)
            test_X = split["test_X"]
            test_y = split["test_y"]
            di_list.append(disparate_impact_scorer(trained_remi, test_X, test_y))
        di = pd.Series(di_list)
        _, _, function_name, _ = traceback.extract_stack()[-2]
        print(f"disparate impact {di.mean():.3f} +- {di.std():.3f} {function_name}")
        self.assertTrue(
            min_di <= di.mean() <= max_di,
            f"{min_di} <= {di.mean()} <= {max_di}",
        )

    def test_adversarial_debiasing_pd_cat(self):
        if tensorflow_installed:
            fairness_info = self.creditg_pd_cat["fairness_info"]
            trainable_remi = AdversarialDebiasing(
                **fairness_info, preparation=self.prep_pd_cat
            )
            self._attempt_remi_creditg_pd_cat(fairness_info, trainable_remi, 0.0, 1.5)

    def test_calibrated_eq_odds_postprocessing_pd_cat(self):
        fairness_info = self.creditg_pd_cat["fairness_info"]
        estim = self.prep_pd_cat >> LogisticRegression(max_iter=1000)
        trainable_remi = CalibratedEqOddsPostprocessing(
            **fairness_info, estimator=estim
        )
        self._attempt_remi_creditg_pd_cat(fairness_info, trainable_remi, 0.65, 0.85)

    def test_disparate_impact_remover_pd_cat(self):
        fairness_info = self.creditg_pd_cat["fairness_info"]
        trainable_remi = DisparateImpactRemover(
            **fairness_info, preparation=self.prep_pd_cat
        ) >> LogisticRegression(max_iter=1000)
        self._attempt_remi_creditg_pd_cat(fairness_info, trainable_remi, 0.72, 0.92)

    def test_disparate_impact_remover_pd_cat_no_redact(self):
        fairness_info = self.creditg_pd_cat["fairness_info"]
        trainable_remi = DisparateImpactRemover(
            **fairness_info, redact=False, preparation=self.prep_pd_cat
        ) >> LogisticRegression(max_iter=1000)
        self._attempt_remi_creditg_pd_cat(fairness_info, trainable_remi, 0.65, 0.75)

    def test_eq_odds_postprocessing_pd_cat(self):
        fairness_info = self.creditg_pd_cat["fairness_info"]
        estim = self.prep_pd_cat >> LogisticRegression(max_iter=1000)
        trainable_remi = EqOddsPostprocessing(**fairness_info, estimator=estim)
        self._attempt_remi_creditg_pd_cat(fairness_info, trainable_remi, 0.82, 1.02)

    def test_gerry_fair_classifier_pd_cat(self):
        fairness_info = self.creditg_pd_cat["fairness_info"]
        trainable_remi = GerryFairClassifier(
            **fairness_info, preparation=self.prep_pd_cat
        )
        self._attempt_remi_creditg_pd_cat(fairness_info, trainable_remi, 0.677, 0.678)

    def test_lfr_pd_cat(self):
        if numba_installed:
            fairness_info = self.creditg_pd_cat["fairness_info"]
            trainable_remi = LFR(
                **fairness_info, preparation=self.prep_pd_cat
            ) >> LogisticRegression(max_iter=1000)
            self._attempt_remi_creditg_pd_cat(fairness_info, trainable_remi, 0.95, 1.05)

    def test_meta_fair_classifier_pd_cat(self):
        if aif360.__version__ >= "4.0.0":
            fairness_info = self.creditg_pd_cat["fairness_info"]
            trainable_remi = MetaFairClassifier(
                **fairness_info, preparation=self.prep_pd_cat
            )
            self._attempt_remi_creditg_pd_cat(fairness_info, trainable_remi, 0.62, 0.87)

    def test_optim_preproc_pd_cat(self):
        # TODO: set the optimizer options as shown in the example https://github.com/Trusted-AI/AIF360/blob/master/examples/demo_optim_data_preproc.ipynb
        if cvxpy_installed:
            fairness_info = self.creditg_pd_cat["fairness_info"]
            _ = OptimPreproc(**fairness_info, optim_options={}) >> LogisticRegression(
                max_iter=1000
            )
            # TODO: this test does not yet call fit or predict

    def test_prejudice_remover_pd_cat(self):
        fairness_info = self.creditg_pd_cat["fairness_info"]
        trainable_remi = PrejudiceRemover(**fairness_info, preparation=self.prep_pd_cat)
        self._attempt_remi_creditg_pd_cat(fairness_info, trainable_remi, 0.70, 0.80)

    def test_redacting_pd_cat(self):
        fairness_info = self.creditg_pd_cat["fairness_info"]
        estim = self.prep_pd_cat >> LogisticRegression(max_iter=1000)
        trainable_remi = Redacting(**fairness_info) >> estim
        self._attempt_remi_creditg_pd_cat(fairness_info, trainable_remi, 0.78, 0.94)

    def test_reject_option_classification_pd_cat(self):
        fairness_info = self.creditg_pd_cat["fairness_info"]
        estim = self.prep_pd_cat >> LogisticRegression(max_iter=1000)
        trainable_remi = RejectOptionClassification(**fairness_info, estimator=estim)
        self._attempt_remi_creditg_pd_cat(fairness_info, trainable_remi, 0.88, 0.98)

    def test_sans_mitigation_pd_cat(self):
        fairness_info = self.creditg_pd_cat["fairness_info"]
        trainable_remi = self.prep_pd_cat >> LogisticRegression(max_iter=1000)
        self._attempt_remi_creditg_pd_cat(fairness_info, trainable_remi, 0.66, 0.76)

    def test_reweighing_pd_cat(self):
        fairness_info = self.creditg_pd_cat["fairness_info"]
        estim = self.prep_pd_cat >> LogisticRegression(max_iter=1000)
        trainable_remi = Reweighing(estimator=estim, **fairness_info)
        self._attempt_remi_creditg_pd_cat(fairness_info, trainable_remi, 0.85, 1.00)

    def test_pd_cat_y_not_series(self):
        fairness_info = self.creditg_pd_cat["fairness_info"]
        trainable_remi = DisparateImpactRemover(
            **fairness_info, preparation=self.prep_pd_cat
        ) >> LogisticRegression(max_iter=1000)
        train_X = self.creditg_pd_cat["splits"][0]["train_X"]
        train_y = self.creditg_pd_cat["splits"][0]["train_y"].to_frame()
        trained_remi = trainable_remi.fit(train_X, train_y)
        test_X = self.creditg_pd_cat["splits"][0]["test_X"]
        test_y = self.creditg_pd_cat["splits"][0]["test_y"].to_frame()
        di_scorer = lale.lib.aif360.disparate_impact(**fairness_info)
        di = di_scorer(trained_remi, test_X, test_y)
        self.assertLessEqual(0.8, di)
        self.assertLessEqual(di, 1.0)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lale.docstrings
import lale.operators
from lale.lib.lale.no_op import NoOp


class _BothImpl:
    # This should be equivalent to:
    # (op1 >> op2) | (op2 >> op1)
    # but with a smaller search space
    def __init__(self, op1, op2, order: str = "forward"):
        self._hyperparams = {"order": order, "op1": op1, "op2": op2}

    def getPipeline(self):
        params = self._hyperparams
        op1 = params.get("op1", None)
        if op1 is None:
            op1 = NoOp()

        op2 = params.get("op2", None)
        if op2 is None:
            op2 = NoOp()

        if params["order"] == "backward":
            return op2 >> op1
        else:
            return op1 >> op2

    def transform(self, X, y=None):
        return self.getPipeline().transform(X, y=y)

    #    def transform_schema(self, s_X):
    #        return self.getPipeline().transform_schema(s_X)

    def predict(self, X, **predict_params):
        return self.getPipeline().predict(X, **predict_params)

    def predict_proba(self, X):
        return self.getPipeline().predict_proba(X)

    def fit(self, X, y=None, **fit_params):
        return self.getPipeline().fit(X, y=y, **fit_params)

    # def get_feature_names(self, input_features=None):
    #     if input_features is not None:
    #         return list(input_features)
    #     elif self._feature_names is not None:
    #         return self._feature_names
    #     else:
    #         raise ValueError('Can only call get_feature_names on a trained operator. Please call fit to get a trained operator.')


_hyperparams_schema = {
    "description": "Hyperparameter schema for the both Higher Order Operator, which wraps another operator and runs it a given number of times",
    "allOf": [
        {
            "description": "This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters",
            "type": "object",
            "additionalProperties": False,
            "relevantToOptimizer": ["order", "op1", "op2"],
            "properties": {
                "order": {"enum": ["forward", "backward"], "default": "forward"},
                "op1": {"laleType": "operator"},
                "op2": {"laleType": "operator"},
            },
        }
    ],
}

# TODO: can we surface the base op input/output schema?
_input_fit_schema = {
    "description": "Input data schema for training both.",
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {"X": {}},
}

_input_predict_transform_schema = (
    {  # TODO: separate predict vs. predict_proba vs. transform
        "description": "Input data schema for transformations using both.",
        "type": "object",
        "required": ["X", "y"],
        "additionalProperties": False,
        "properties": {"X": {}, "y": {}},
    }
)

_output_schema = {  # TODO: separate predict vs. predict_proba vs. transform
    "description": "Output data schema for transformations using both.",
    "laleType": "Any",
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.lale.both.html",
    "import_from": "lale.lib.lale",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_transform_schema,
        "output_predict": _output_schema,
        "input_predict_proba": _input_predict_transform_schema,
        "output_predict_proba": _output_schema,
        "input_transform": _input_predict_transform_schema,
        "output_transform": _output_schema,
    },
}


Both = lale.operators.make_operator(_BothImpl, _combined_schemas)

lale.docstrings.set_docstrings(Both)



from numpy import inf, nan
from sklearn.naive_bayes import MultinomialNB as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _MultinomialNBImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

    def predict_proba(self, X):
        return self._wrapped_model.predict_proba(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for MultinomialNB    Naive Bayes classifier for multinomial models",
    "allOf": [
        {
            "type": "object",
            "required": ["alpha", "fit_prior", "class_prior"],
            "relevantToOptimizer": ["alpha", "fit_prior"],
            "additionalProperties": False,
            "properties": {
                "alpha": {
                    "type": "number",
                    "minimumForOptimizer": 1e-10,
                    "maximumForOptimizer": 1.0,
                    "distribution": "loguniform",
                    "default": 1.0,
                    "description": "Additive (Laplace/Lidstone) smoothing parameter (0 for no smoothing).",
                },
                "fit_prior": {
                    "type": "boolean",
                    "default": True,
                    "description": "Whether to learn class prior probabilities or not",
                },
                "class_prior": {
                    "anyOf": [
                        {"type": "array", "items": {"type": "number"}},
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "Prior probabilities of the classes",
                },
            },
        }
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Fit Naive Bayes classifier according to X, y",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "Training vectors, where n_samples is the number of samples and n_features is the number of features.",
        },
        "y": {
            "type": "array",
            "items": {"type": "number"},
            "description": "Target values.",
        },
        "sample_weight": {
            "anyOf": [{"type": "array", "items": {"type": "number"}}, {"enum": [None]}],
            "default": None,
            "description": "Weights applied to individual samples (1",
        },
    },
}
_input_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Perform classification on an array of test vectors X.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {"type": "array", "items": {"type": "array", "items": {"type": "number"}}}
    },
}
_output_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predicted target values for X",
    "type": "array",
    "items": {"type": "number"},
}
_input_predict_proba_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Return probability estimates for the test vector X.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {"type": "array", "items": {"type": "array", "items": {"type": "number"}}}
    },
}
_output_predict_proba_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Returns the probability of the samples for each class in the model",
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.naive_bayes.MultinomialNB#sklearn-naive_bayes-multinomialnb",
    "import_from": "sklearn.naive_bayes",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
        "input_predict_proba": _input_predict_proba_schema,
        "output_predict_proba": _output_predict_proba_schema,
    },
}
MultinomialNB = make_operator(_MultinomialNBImpl, _combined_schemas)

set_docstrings(MultinomialNB)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from test import EnableSchemaValidation

import jsonschema
import numpy as np
import sklearn
import sklearn.datasets
import sklearn.metrics
import sklearn.model_selection
import xgboost

from lale.lib.lale import Hyperopt
from lale.lib.sklearn import (
    SVC,
    DecisionTreeClassifier,
    DecisionTreeRegressor,
    ExtraTreesClassifier,
    ExtraTreesRegressor,
    FeatureAgglomeration,
    FunctionTransformer,
    GradientBoostingClassifier,
    GradientBoostingRegressor,
    LinearRegression,
    LogisticRegression,
    MLPClassifier,
    PolynomialFeatures,
    RandomForestClassifier,
    RandomForestRegressor,
    Ridge,
    VotingClassifier,
)
from lale.lib.xgboost import XGBClassifier, XGBRegressor

assert sklearn.__version__ == "0.20.3", "This test is for scikit-learn 0.20.3."
assert xgboost.__version__ == "0.90", "This test is for XGBoost 0.90."


class TestDecisionTreeClassifier(unittest.TestCase):
    def setUp(self):
        X, y = sklearn.datasets.load_iris(return_X_y=True)
        (
            self.train_X,
            self.test_X,
            self.train_y,
            self.test_y,
        ) = sklearn.model_selection.train_test_split(X, y)

    def test_with_defaults(self):
        trainable = DecisionTreeClassifier()
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_ccp_alpha(self):
        with EnableSchemaValidation():
            with self.assertRaisesRegex(
                jsonschema.ValidationError, "argument 'ccp_alpha' was unexpected"
            ):
                _ = DecisionTreeClassifier(ccp_alpha=0.01)

    def test_with_hyperopt(self):
        planned = DecisionTreeClassifier
        trained = planned.auto_configure(
            self.train_X, self.train_y, optimizer=Hyperopt, cv=3, max_evals=3
        )
        _ = trained.predict(self.test_X)


class TestDecisionTreeRegressor(unittest.TestCase):
    def setUp(self):
        X, y = sklearn.datasets.load_diabetes(return_X_y=True)
        (
            self.train_X,
            self.test_X,
            self.train_y,
            self.test_y,
        ) = sklearn.model_selection.train_test_split(X, y)

    def test_with_defaults(self):
        trainable = DecisionTreeRegressor()
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_ccp_alpha(self):
        with EnableSchemaValidation():
            with self.assertRaisesRegex(
                jsonschema.ValidationError, "argument 'ccp_alpha' was unexpected"
            ):
                _ = DecisionTreeRegressor(ccp_alpha=0.01)

    def test_with_hyperopt(self):
        planned = DecisionTreeRegressor
        trained = planned.auto_configure(
            self.train_X,
            self.train_y,
            optimizer=Hyperopt,
            scoring="r2",
            cv=3,
            max_evals=3,
        )
        _ = trained.predict(self.test_X)


class TestExtraTreesClassifier(unittest.TestCase):
    def setUp(self):
        X, y = sklearn.datasets.load_iris(return_X_y=True)
        (
            self.train_X,
            self.test_X,
            self.train_y,
            self.test_y,
        ) = sklearn.model_selection.train_test_split(X, y)

    def test_with_defaults(self):
        trainable = ExtraTreesClassifier()
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_n_estimators(self):
        default = ExtraTreesClassifier.get_defaults()["n_estimators"]
        self.assertEqual(default, 10)

    def test_ccp_alpha(self):
        with EnableSchemaValidation():
            with self.assertRaisesRegex(
                jsonschema.ValidationError, "argument 'ccp_alpha' was unexpected"
            ):
                _ = ExtraTreesClassifier(ccp_alpha=0.01)

    def test_max_samples(self):
        with EnableSchemaValidation():
            with self.assertRaisesRegex(
                jsonschema.ValidationError, "argument 'max_samples' was unexpected"
            ):
                _ = ExtraTreesClassifier(max_samples=0.01)

    def test_with_hyperopt(self):
        planned = ExtraTreesClassifier
        trained = planned.auto_configure(
            self.train_X, self.train_y, optimizer=Hyperopt, cv=3, max_evals=3
        )
        _ = trained.predict(self.test_X)


class TestExtraTreesRegressor(unittest.TestCase):
    def setUp(self):
        X, y = sklearn.datasets.load_diabetes(return_X_y=True)
        (
            self.train_X,
            self.test_X,
            self.train_y,
            self.test_y,
        ) = sklearn.model_selection.train_test_split(X, y)

    def test_with_defaults(self):
        trainable = ExtraTreesRegressor()
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_n_estimators(self):
        default = ExtraTreesRegressor.get_defaults()["n_estimators"]
        self.assertEqual(default, 10)

    def test_ccp_alpha(self):
        with EnableSchemaValidation():
            with self.assertRaisesRegex(
                jsonschema.ValidationError, "argument 'ccp_alpha' was unexpected"
            ):
                _ = ExtraTreesRegressor(ccp_alpha=0.01)

    def test_max_samples(self):
        with EnableSchemaValidation():
            with self.assertRaisesRegex(
                jsonschema.ValidationError, "argument 'max_samples' was unexpected"
            ):
                _ = ExtraTreesRegressor(max_samples=0.01)

    def test_with_hyperopt(self):
        planned = ExtraTreesRegressor
        trained = planned.auto_configure(
            self.train_X,
            self.train_y,
            scoring="r2",
            optimizer=Hyperopt,
            cv=3,
            max_evals=3,
        )
        _ = trained.predict(self.test_X)


class TestFeatureAgglomeration(unittest.TestCase):
    def setUp(self):
        X, y = sklearn.datasets.load_iris(return_X_y=True)
        (
            self.train_X,
            self.test_X,
            self.train_y,
            self.test_y,
        ) = sklearn.model_selection.train_test_split(X, y)

    def test_with_defaults(self):
        trainable = FeatureAgglomeration() >> LogisticRegression()
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_with_hyperopt(self):
        planned = FeatureAgglomeration >> LogisticRegression
        trained = planned.auto_configure(
            self.train_X,
            self.train_y,
            optimizer=Hyperopt,
            cv=3,
            max_evals=3,
            verbose=True,
        )
        _ = trained.predict(self.test_X)


class TestFunctionTransformer(unittest.TestCase):
    def setUp(self):
        X, y = sklearn.datasets.load_iris(return_X_y=True)
        (
            self.train_X,
            self.test_X,
            self.train_y,
            self.test_y,
        ) = sklearn.model_selection.train_test_split(X, y)

    def test_with_defaults(self):
        trainable = FunctionTransformer(func=np.log1p) >> LogisticRegression()
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_pass_y(self):
        trainable = (
            FunctionTransformer(func=np.log1p, pass_y=False) >> LogisticRegression()
        )
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_validate(self):
        default = FunctionTransformer.get_defaults()["validate"]
        self.assertEqual(default, True)

    def test_with_hyperopt(self):
        planned = FunctionTransformer(func=np.log1p) >> LogisticRegression
        trained = planned.auto_configure(
            self.train_X, self.train_y, optimizer=Hyperopt, cv=3, max_evals=3
        )
        _ = trained.predict(self.test_X)


class TestGradientBoostingClassifier(unittest.TestCase):
    def setUp(self):
        X, y = sklearn.datasets.load_iris(return_X_y=True)
        (
            self.train_X,
            self.test_X,
            self.train_y,
            self.test_y,
        ) = sklearn.model_selection.train_test_split(X, y)

    def test_with_defaults(self):
        trainable = GradientBoostingClassifier()
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_ccp_alpha(self):
        with EnableSchemaValidation():
            with self.assertRaisesRegex(
                jsonschema.ValidationError, "argument 'ccp_alpha' was unexpected"
            ):
                _ = GradientBoostingClassifier(ccp_alpha=0.01)

    def test_with_hyperopt(self):
        planned = GradientBoostingClassifier
        trained = planned.auto_configure(
            self.train_X, self.train_y, optimizer=Hyperopt, cv=3, max_evals=3
        )
        _ = trained.predict(self.test_X)


class TestGradientBoostingRegressor(unittest.TestCase):
    def setUp(self):
        X, y = sklearn.datasets.load_diabetes(return_X_y=True)
        (
            self.train_X,
            self.test_X,
            self.train_y,
            self.test_y,
        ) = sklearn.model_selection.train_test_split(X, y)

    def test_with_defaults(self):
        trainable = GradientBoostingRegressor()
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_ccp_alpha(self):
        with EnableSchemaValidation():
            with self.assertRaisesRegex(
                jsonschema.ValidationError, "argument 'ccp_alpha' was unexpected"
            ):
                _ = GradientBoostingRegressor(ccp_alpha=0.01)

    def test_with_hyperopt(self):
        planned = GradientBoostingRegressor
        trained = planned.auto_configure(
            self.train_X,
            self.train_y,
            scoring="r2",
            optimizer=Hyperopt,
            cv=3,
            max_evals=3,
        )
        _ = trained.predict(self.test_X)


class TestLinearRegression(unittest.TestCase):
    def setUp(self):
        X, y = sklearn.datasets.load_diabetes(return_X_y=True)
        (
            self.train_X,
            self.test_X,
            self.train_y,
            self.test_y,
        ) = sklearn.model_selection.train_test_split(X, y)

    def test_with_defaults(self):
        trainable = LinearRegression()
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_with_hyperopt(self):
        planned = LinearRegression
        trained = planned.auto_configure(
            self.train_X,
            self.train_y,
            scoring="r2",
            optimizer=Hyperopt,
            cv=3,
            max_evals=3,
        )
        _ = trained.predict(self.test_X)


class TestLogisticRegression(unittest.TestCase):
    def setUp(self):
        X, y = sklearn.datasets.load_iris(return_X_y=True)
        (
            self.train_X,
            self.test_X,
            self.train_y,
            self.test_y,
        ) = sklearn.model_selection.train_test_split(X, y)

    def test_with_defaults(self):
        trainable = LogisticRegression()
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_multi_class(self):
        default = LogisticRegression.get_defaults()["multi_class"]
        self.assertEqual(default, "ovr")

    def test_solver(self):
        default = LogisticRegression.get_defaults()["solver"]
        self.assertEqual(default, "liblinear")

    def test_l1_ratio(self):
        with EnableSchemaValidation():
            with self.assertRaisesRegex(
                jsonschema.ValidationError, "argument 'l1_ratio' was unexpected"
            ):
                _ = LogisticRegression(l1_ratio=0.2)

    def test_with_hyperopt(self):
        planned = LogisticRegression
        trained = planned.auto_configure(
            self.train_X, self.train_y, optimizer=Hyperopt, cv=3, max_evals=3
        )
        _ = trained.predict(self.test_X)


class TestMLPClassifier(unittest.TestCase):
    def setUp(self):
        X, y = sklearn.datasets.load_iris(return_X_y=True)
        (
            self.train_X,
            self.test_X,
            self.train_y,
            self.test_y,
        ) = sklearn.model_selection.train_test_split(X, y)

    def test_with_defaults(self):
        trainable = MLPClassifier()
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_max_fun(self):
        with EnableSchemaValidation():
            with self.assertRaisesRegex(
                jsonschema.ValidationError, "argument 'max_fun' was unexpected"
            ):
                _ = MLPClassifier(max_fun=1000)

    def test_with_hyperopt(self):
        planned = MLPClassifier(max_iter=20)
        trained = planned.auto_configure(
            self.train_X, self.train_y, optimizer=Hyperopt, cv=3, max_evals=3
        )
        _ = trained.predict(self.test_X)


class TestPolynomialFeatures(unittest.TestCase):
    def setUp(self):
        X, y = sklearn.datasets.load_iris(return_X_y=True)
        (
            self.train_X,
            self.test_X,
            self.train_y,
            self.test_y,
        ) = sklearn.model_selection.train_test_split(X, y)

    def test_with_defaults(self):
        trainable = PolynomialFeatures() >> LogisticRegression()
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_order(self):
        with EnableSchemaValidation():
            with self.assertRaisesRegex(
                jsonschema.ValidationError, "argument 'order' was unexpected"
            ):
                _ = PolynomialFeatures(order="F") >> LogisticRegression()

    def test_with_hyperopt(self):
        planned = PolynomialFeatures >> LogisticRegression
        trained = planned.auto_configure(
            self.train_X, self.train_y, optimizer=Hyperopt, cv=3, max_evals=3
        )
        _ = trained.predict(self.test_X)


class TestRandomForestClassifier(unittest.TestCase):
    def setUp(self):
        X, y = sklearn.datasets.load_iris(return_X_y=True)
        (
            self.train_X,
            self.test_X,
            self.train_y,
            self.test_y,
        ) = sklearn.model_selection.train_test_split(X, y)

    def test_with_defaults(self):
        trainable = RandomForestClassifier()
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_n_estimators(self):
        default = RandomForestClassifier.get_defaults()["n_estimators"]
        self.assertEqual(default, 10)

    def test_ccp_alpha(self):
        with EnableSchemaValidation():
            with self.assertRaisesRegex(
                jsonschema.ValidationError, "argument 'ccp_alpha' was unexpected"
            ):
                _ = RandomForestClassifier(ccp_alpha=0.01)

    def test_max_samples(self):
        with EnableSchemaValidation():
            with self.assertRaisesRegex(
                jsonschema.ValidationError, "argument 'max_samples' was unexpected"
            ):
                _ = RandomForestClassifier(max_samples=0.01)

    def test_with_hyperopt(self):
        planned = RandomForestClassifier
        trained = planned.auto_configure(
            self.train_X, self.train_y, optimizer=Hyperopt, cv=3, max_evals=3
        )
        _ = trained.predict(self.test_X)


class TestRandomForestRegressor(unittest.TestCase):
    def setUp(self):
        X, y = sklearn.datasets.load_diabetes(return_X_y=True)
        (
            self.train_X,
            self.test_X,
            self.train_y,
            self.test_y,
        ) = sklearn.model_selection.train_test_split(X, y)

    def test_with_defaults(self):
        trainable = RandomForestRegressor()
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_n_estimators(self):
        default = RandomForestRegressor.get_defaults()["n_estimators"]
        self.assertEqual(default, 10)

    def test_ccp_alpha(self):
        with EnableSchemaValidation():
            with self.assertRaisesRegex(
                jsonschema.ValidationError, "argument 'ccp_alpha' was unexpected"
            ):
                _ = RandomForestRegressor(ccp_alpha=0.01)

    def test_max_samples(self):
        with self.assertRaisesRegex(
            jsonschema.ValidationError, "argument 'max_samples' was unexpected"
        ):
            _ = RandomForestRegressor(max_samples=0.01)

    def test_with_hyperopt(self):
        planned = RandomForestRegressor
        trained = planned.auto_configure(
            self.train_X,
            self.train_y,
            scoring="r2",
            optimizer=Hyperopt,
            cv=3,
            max_evals=3,
        )
        _ = trained.predict(self.test_X)


class TestRidge(unittest.TestCase):
    def setUp(self):
        X, y = sklearn.datasets.load_diabetes(return_X_y=True)
        (
            self.train_X,
            self.test_X,
            self.train_y,
            self.test_y,
        ) = sklearn.model_selection.train_test_split(X, y)

    def test_with_defaults(self):
        trainable = Ridge()
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_with_hyperopt(self):
        planned = Ridge
        trained = planned.auto_configure(
            self.train_X,
            self.train_y,
            scoring="r2",
            optimizer=Hyperopt,
            cv=3,
            max_evals=3,
        )
        _ = trained.predict(self.test_X)


class TestSVC(unittest.TestCase):
    def setUp(self):
        X, y = sklearn.datasets.load_iris(return_X_y=True)
        (
            self.train_X,
            self.test_X,
            self.train_y,
            self.test_y,
        ) = sklearn.model_selection.train_test_split(X, y)

    def test_with_defaults(self):
        trainable = SVC()
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_gamma(self):
        default = SVC.get_defaults()["gamma"]
        self.assertEqual(default, "auto_deprecated")

    def test_break_ties(self):
        with EnableSchemaValidation():
            with self.assertRaisesRegex(
                jsonschema.ValidationError, "argument 'break_ties' was unexpected"
            ):
                _ = SVC(break_ties=True)

    def test_with_hyperopt(self):
        planned = SVC
        trained = planned.auto_configure(
            self.train_X, self.train_y, optimizer=Hyperopt, cv=3, max_evals=3
        )
        _ = trained.predict(self.test_X)


class TestVotingClassifier(unittest.TestCase):
    def setUp(self):
        X, y = sklearn.datasets.load_iris(return_X_y=True)
        (
            self.train_X,
            self.test_X,
            self.train_y,
            self.test_y,
        ) = sklearn.model_selection.train_test_split(X, y)

    def test_with_defaults(self):
        trainable = VotingClassifier(
            estimators=[("lr", LogisticRegression()), ("dt", DecisionTreeClassifier())]
        )
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_estimators(self):
        trainable = VotingClassifier(
            estimators=[
                ("lr", LogisticRegression()),
                ("dt", DecisionTreeClassifier()),
                ("na", None),
            ]
        )
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_with_hyperopt(self):
        planned = VotingClassifier(
            estimators=[("lr", LogisticRegression), ("dt", DecisionTreeClassifier)]
        )
        trained = planned.auto_configure(
            self.train_X, self.train_y, optimizer=Hyperopt, cv=3, max_evals=3
        )
        _ = trained.predict(self.test_X)


class TestXGBClassifier(unittest.TestCase):
    def setUp(self):
        X, y = sklearn.datasets.load_iris(return_X_y=True)
        (
            self.train_X,
            self.test_X,
            self.train_y,
            self.test_y,
        ) = sklearn.model_selection.train_test_split(X, y)

    def test_with_defaults(self):
        trainable = XGBClassifier()
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_with_hyperopt(self):
        planned = XGBClassifier
        trained = planned.auto_configure(
            self.train_X, self.train_y, optimizer=Hyperopt, cv=3, max_evals=3
        )
        _ = trained.predict(self.test_X)


class TestXGBRegressor(unittest.TestCase):
    def setUp(self):
        X, y = sklearn.datasets.load_diabetes(return_X_y=True)
        (
            self.train_X,
            self.test_X,
            self.train_y,
            self.test_y,
        ) = sklearn.model_selection.train_test_split(X, y)

    def test_with_defaults(self):
        trainable = XGBRegressor()
        trained = trainable.fit(self.train_X, self.train_y)
        _ = trained.predict(self.test_X)

    def test_with_hyperopt(self):
        planned = XGBRegressor
        trained = planned.auto_configure(
            self.train_X,
            self.train_y,
            scoring="r2",
            optimizer=Hyperopt,
            cv=3,
            max_evals=3,
        )
        _ = trained.predict(self.test_X)



# Copyright 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numbers
import typing

import numpy as np
import pandas as pd

import lale.docstrings
import lale.operators
from lale.expressions import it, mean, median, mode, replace
from lale.helpers import _is_df, _is_pandas_df, _is_spark_df
from lale.lib.sklearn import simple_imputer
from lale.schemas import Enum

from .aggregate import Aggregate
from .map import Map


def _is_numeric_df(X):
    if _is_pandas_df(X):
        return X.shape[1] == X.select_dtypes(include=np.number).shape[1]
    elif _is_spark_df(X):
        from pyspark.sql.types import NumericType

        numeric_cols = [
            f.name for f in X.schema.fields if isinstance(f.dataType, NumericType)
        ]
        return len(X.columns) == len(numeric_cols)
    else:
        return False


def _is_string_df(X):
    if _is_pandas_df(X):
        return X.shape[1] == X.select_dtypes(include="object").shape[1]
    elif _is_spark_df(X):
        from pyspark.sql.types import StringType

        numeric_cols = [
            f.name for f in X.schema.fields if isinstance(f.dataType, StringType)
        ]
        return len(X.columns) == len(numeric_cols)
    else:
        return False


class _SimpleImputerImpl:
    def __init__(
        self,
        missing_values=np.nan,
        strategy="mean",
        fill_value=None,
        verbose=0,
        copy=True,
        add_indicator=False,
    ):

        self.missing_values = missing_values
        self.strategy = strategy
        self.fill_value = fill_value
        self.verbose = verbose
        if not copy:
            raise ValueError("This implementation only supports `copy=True`.")

        if add_indicator:
            raise ValueError("This implementation only supports `add_indicator=False`.")
        # the `indicator_`` property is always None as we do not support `add_indictor=True`
        self.indicator_ = None

    def fit(self, X, y=None):

        self._validate_input(X)

        # assign appropriate value to fill_value depending on the datatype.
        # default fill_value is 0 for numerical input and "missing_value"
        # otherwise
        if self.fill_value is None:
            if _is_numeric_df(X):
                fill_value = 0
            else:
                fill_value = "missing_value"
        else:
            fill_value = self.fill_value

        # validate that fill_value is numerical for numerical data
        if (
            self.strategy == "constant"
            and _is_numeric_df(X)
            and not isinstance(fill_value, numbers.Real)
        ):
            raise ValueError(
                "'fill_value'={0} is invalid. Expected a "
                "numerical value when imputing numerical "
                "data".format(fill_value)
            )

        # set attribute values
        self.n_features_in_ = len(X.columns)
        self.feature_names_in_ = X.columns

        agg_op = None
        agg_data = None
        # learn the values to be imputed
        if self.strategy == "mean":
            agg_op = Aggregate(
                columns={c: mean(it[c]) for c in X.columns},
                exclude_value=self.missing_values,
            )
        elif self.strategy == "median":
            agg_op = Aggregate(
                columns={c: median(it[c]) for c in X.columns},
                exclude_value=self.missing_values,
            )
        elif self.strategy == "most_frequent":
            agg_op = Aggregate(
                columns={c: mode(it[c]) for c in X.columns},
                exclude_value=self.missing_values,
            )
        elif self.strategy == "constant":
            agg_data = [[fill_value for col in X.columns]]
            agg_data = pd.DataFrame(agg_data, columns=X.columns)
        if agg_data is None and agg_op is not None:
            agg_data = agg_op.transform(X)
        if agg_data is not None and _is_spark_df(agg_data):
            agg_data = agg_data.toPandas()

        if agg_data is not None and _is_pandas_df(agg_data):
            self.statistics_ = agg_data.to_numpy()[
                0
            ]  # Converting from a 2-d array to 1-d
        # prepare the transformer
        if agg_data is not None:
            self.transformer = Map(
                columns={
                    col_name: replace(
                        it[col_name], {self.missing_values: agg_data.iloc[0, col_idx]}
                    )
                    for col_idx, col_name in enumerate(X.columns)
                }
            )
        return self

    def transform(self, X):
        return self.transformer.transform(X)

    def _validate_input(self, X):
        # validate that the dataset is either a pandas dataframe or spark.
        # For example, sparse matrix is not allowed.
        if not _is_df(X):
            raise ValueError(
                f"""Unsupported type(X) {type(X)} for SimpleImputer.
            Only pandas.DataFrame or pyspark.sql.DataFrame are allowed."""
            )
        # validate input to check the correct dtype and strategy
        # `mean` and `median` are not applicable to string inputs
        if not _is_numeric_df(X) and self.strategy in ["mean", "median"]:
            raise ValueError(
                "Cannot use {} strategy with non-numeric data.".format(self.strategy)
            )

        # Check that missing_values are the right type
        if _is_numeric_df(X) and not isinstance(self.missing_values, numbers.Real):
            raise ValueError(
                "'X' and 'missing_values' types are expected to be"
                " both numerical. Got X.dtypes={} and "
                " type(missing_values)={}.".format(X.dtypes, type(self.missing_values))
            )


_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """Relational algebra reimplementation of scikit-learn's `SimpleImputer`_.
Works on both pandas and Spark dataframes by using `Aggregate`_ for `fit` and `Map`_ for `transform`, which in turn use the appropriate backend.

.. _`SimpleImputer`: https://scikit-learn.org/stable/modules/generated/sklearn.imputer.SimpleImputer.html
.. _`Aggregate`: https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.aggregate.html
.. _`Map`: https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.map.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.simple_imputer.html",
    "type": "object",
    "tags": {
        "pre": [],
        "op": ["transformer", "interpretable"],
        "post": [],
    },
    "properties": {
        "hyperparams": simple_imputer._hyperparams_schema,
        "input_fit": simple_imputer._input_fit_schema,
        "input_transform": simple_imputer._input_transform_schema,
        "output_transform": simple_imputer._output_transform_schema,
    },
}

SimpleImputer = lale.operators.make_operator(_SimpleImputerImpl, _combined_schemas)

SimpleImputer = typing.cast(
    lale.operators.PlannedIndividualOp,
    SimpleImputer.customize_schema(
        copy=Enum(
            values=[True],
            desc="`copy=True` is the only value currently supported by this implementation",
            default=True,
        ),
        add_indicator=Enum(
            values=[False],
            desc="`add_indicator=False` is the only value currently supported by this implementation",
            default=False,
        ),
    ),
)

lale.docstrings.set_docstrings(SimpleImputer)



# Copyright 2020 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd

import lale.datasets.data_schemas
import lale.docstrings
import lale.operators
from lale.helpers import (
    _get_subscript_value,
    _is_ast_attribute,
    _is_ast_subscript,
    _is_df,
    _is_spark_df,
)

try:
    from pyspark.sql.functions import col

    spark_installed = True

except ImportError:
    spark_installed = False


class _JoinImpl:
    def __init__(
        self,
        *,
        pred=None,
        join_limit=None,
        sliding_window_length=None,
        join_type="inner",
        name=None,
    ):
        self.pred = pred
        self.join_type = join_type
        self.name = name

    # Parse the predicate element passed as input
    @classmethod
    def _get_join_info(cls, expr_to_parse):
        left_key = []
        right_key = []
        if _is_ast_subscript(expr_to_parse.left.value):
            left_name = _get_subscript_value(expr_to_parse.left.value)
        elif _is_ast_attribute(expr_to_parse.left.value):
            left_name = expr_to_parse.left.value.attr
        else:
            raise ValueError(
                "ERROR: Expression type not supported! Formats supported: it.table_name.column_name or it['table_name'].column_name"
            )
        if _is_ast_subscript(expr_to_parse.left):
            left_key.append(_get_subscript_value(expr_to_parse.left))
        elif _is_ast_attribute(expr_to_parse.left):
            left_key.append(expr_to_parse.left.attr)
        else:
            raise ValueError(
                "ERROR: Expression type not supported! Formats supported: it.table_name.column_name or it.table_name['column_name']"
            )
        if _is_ast_subscript(expr_to_parse.comparators[0].value):
            right_name = _get_subscript_value(expr_to_parse.comparators[0].value)
        elif _is_ast_attribute(expr_to_parse.comparators[0].value):
            right_name = expr_to_parse.comparators[0].value.attr
        else:
            raise ValueError(
                "ERROR: Expression type not supported! Formats supported: it.table_name.column_name or it['table_name'].column_name"
            )
        if _is_ast_subscript(expr_to_parse.comparators[0]):
            right_key.append(_get_subscript_value(expr_to_parse.comparators[0]))
        elif _is_ast_attribute(expr_to_parse.comparators[0]):
            right_key.append(expr_to_parse.comparators[0].attr)
        else:
            raise ValueError(
                "ERROR: Expression type not supported! Formats supported: it.table_name.column_name or it.table_name['column_name']"
            )
        return left_name, left_key, right_name, right_key

    @classmethod
    def validate_hyperparams(cls, pred=None, **hyperparams):
        tables_encountered = set()

        for key in pred:
            if isinstance(key, list):
                sub_list_tables = (
                    list()
                )  # use an ordered list to improve error messages
                for sub_key in key:
                    (
                        left_table_name,
                        left_key_col,
                        right_table_name,
                        right_key_col,
                    ) = cls._get_join_info(sub_key._expr)
                    if sub_list_tables and not (
                        left_table_name in sub_list_tables
                        and right_table_name in sub_list_tables
                    ):
                        sub_list_tables.append(left_table_name)
                        first_table_names = ", ".join(sub_list_tables)
                        raise ValueError(
                            "ERROR: Composite key involving the {}, and {} tables is problematic, since it references more than two tables.".format(
                                first_table_names, right_table_name
                            )
                        )
                    elif tables_encountered and not (
                        left_table_name in tables_encountered
                        or right_table_name in tables_encountered
                    ):
                        left_expr = f"it.{left_table_name}{left_key_col}"
                        right_expr = f"it.{right_table_name}{right_key_col}"
                        raise ValueError(
                            "ERROR: Composite key involving {} == {} is problematic, since neither the {} nor the {} tables were used in a previous key. Join operations must be chained (they can't have two disconnected join conditions)".format(
                                left_expr, right_expr, left_table_name, right_table_name
                            )
                        )
                    sub_list_tables.append(left_table_name)
                    sub_list_tables.append(right_table_name)
                    tables_encountered.add(left_table_name)
                    tables_encountered.add(right_table_name)
            else:
                (
                    left_table_name,
                    left_key_col,
                    right_table_name,
                    right_key_col,
                ) = cls._get_join_info(key._expr)
                if tables_encountered and not (
                    left_table_name in tables_encountered
                    or right_table_name in tables_encountered
                ):
                    left_expr = f"it.{left_table_name}{left_key_col}"
                    right_expr = f"it.{right_table_name}{right_key_col}"
                    raise ValueError(
                        "ERROR: Single key involving {} == {} is problematic, since neither the {} nor the {} tables were used in a previous key. Join operations must be chained (they can't have two disconnected join conditions)".format(
                            left_expr, right_expr, left_table_name, right_table_name
                        )
                    )
                tables_encountered.add(left_table_name)
                tables_encountered.add(right_table_name)

    def transform(self, X):
        # X is assumed to be a list of datasets with get_table_name(d) != None
        joined_df = pd.DataFrame()
        tables_encountered = set()

        # Implementation of join operator
        def join_df(left_df, right_df):

            # Joining spark dataframes
            if _is_spark_df(left_df) and _is_spark_df(right_df):
                on = []
                drop_col = []
                left_table = left_df.alias("left_table")
                right_table = right_df.alias("right_table")

                for k, key in enumerate(left_key_col):
                    on.append(
                        col("{}.{}".format("left_table", key)).eqNullSafe(
                            col("{}.{}".format("right_table", right_key_col[k]))
                        )
                    )
                    if key == right_key_col[k]:
                        drop_col.append(key)
                op_df = left_table.join(right_table, on, self.join_type)
                for key in drop_col:
                    op_df = op_df.drop(getattr(right_table, key))
                return op_df

            # Joining pandas dataframes
            op_df = pd.merge(
                left_df,
                right_df,
                how=self.join_type,
                left_on=left_key_col,
                right_on=right_key_col,
            )
            return op_df

        def fetch_one_df(named_df, table_name):
            if lale.datasets.data_schemas.get_table_name(named_df) == table_name:
                return named_df
            return None

        def fetch_df(left_table_name, right_table_name):
            left_df = []
            right_df = []
            for named_df in X:
                if not tables_encountered:
                    left_df_candidate = fetch_one_df(named_df, left_table_name)
                    if _is_df(left_df_candidate):
                        left_df = left_df_candidate
                    right_df_candidate = fetch_one_df(named_df, right_table_name)
                    if _is_df(right_df_candidate):
                        right_df = right_df_candidate
                else:
                    if left_table_name in tables_encountered:
                        left_df = joined_df
                        right_df_candidate = fetch_one_df(named_df, right_table_name)
                        if _is_df(right_df_candidate):
                            right_df = right_df_candidate
                    elif right_table_name in tables_encountered:
                        right_df = joined_df
                        left_df_candidate = fetch_one_df(named_df, left_table_name)
                        if _is_df(left_df_candidate):
                            left_df = left_df_candidate
            return left_df, right_df

        # Iterate over all the elements of the predicate
        for pred_element in self.pred if self.pred is not None else []:
            left_table_name = ""
            left_key_col = []
            right_table_name = ""
            right_key_col = []
            if isinstance(pred_element, list):
                # Prepare composite key to apply join once for all the participating columns together
                for sub_pred_element in pred_element:
                    (
                        left_table_name,
                        temp_left_key,
                        right_table_name,
                        temp_right_key,
                    ) = self._get_join_info(sub_pred_element._expr)
                    left_key_col.extend(temp_left_key)
                    right_key_col.extend(temp_right_key)
            else:
                (
                    left_table_name,
                    left_key_col,
                    right_table_name,
                    right_key_col,
                ) = self._get_join_info(pred_element._expr)
            left_df, right_df = fetch_df(left_table_name, right_table_name)
            if not _is_df(left_df) or not _is_df(right_df):
                raise ValueError(
                    "ERROR: Cannot perform join operation, either '{}' or '{}' table not present in input X!".format(
                        left_table_name, right_table_name
                    )
                )
            columns_in_both_tables = set(left_df.columns).intersection(  # type: ignore
                set(right_df.columns)  # type: ignore
            )
            if columns_in_both_tables and not set(
                sorted(columns_in_both_tables)
            ) == set(sorted(left_key_col + right_key_col)):
                raise ValueError(
                    "Cannot perform join operation! Non-key columns cannot be duplicate."
                )
            joined_df = join_df(left_df, right_df)
            tables_encountered.add(left_table_name)
            tables_encountered.add(right_table_name)
        return lale.datasets.data_schemas.add_table_name(joined_df, self.name)

    def viz_label(self) -> str:
        if isinstance(self.name, str):
            return f"Join:\n{self.name}"
        return "Join"


_hyperparams_schema = {
    "allOf": [
        {
            "description": "This first sub-object lists all constructor arguments with their "
            "types, one at a time, omitting cross-argument constraints, if any.",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "pred",
                "join_limit",
                "sliding_window_length",
                "join_type",
                "name",
            ],
            "relevantToOptimizer": [],
            "properties": {
                "pred": {
                    "description": "Join predicate. Given as Python AST expression.",
                    "laleType": "Any",
                },
                "join_limit": {
                    "description": """Not yet implemented!
For join paths that are one-to-many, join_limit is use to sample the joined results.
When the right hand side of the join has a timestamp column, the join_limit is applied to select the most recent rows.
When the right hand side does not have a timestamp, it randomly samples join_limit number of rows.
Sampling is applied after each pair of tables are joined.""",
                    "anyOf": [{"type": "number"}, {"enum": [None]}],
                    "default": None,
                },
                "sliding_window_length": {
                    "description": """Not yet implemented!
sliding_window_length is also used for sampling the joined results,
only rows in a recent window of length sliding_window_length seconds is used in addition to join_limit.""",
                    "anyOf": [{"type": "number"}, {"enum": [None]}],
                    "default": None,
                },
                "join_type": {
                    "description": """There are various types of SQL joins available and join_type gives the user the option
to choose which type of join the user wants to implement.""",
                    "enum": ["inner", "left", "right"],
                    "default": "inner",
                },
                "name": {
                    "description": "The table name to be given to the output dataframe.",
                    "anyOf": [
                        {
                            "type": "string",
                            "pattern": "[^ ]",
                            "description": "String (cannot be all spaces).",
                        },
                        {
                            "enum": [None],
                            "description": "No table name.",
                        },
                    ],
                    "default": None,
                },
            },
        }
    ]
}

_input_transform_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "List of tables.",
            "type": "array",
            "items": {"type": "array", "items": {"laleType": "Any"}},
            "minItems": 1,
        }
    },
}

_output_transform_schema = {
    "description": "Features; no restrictions on data type.",
    "laleType": "Any",
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Relational algebra join operator.",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.lale.join.html",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}


Join = lale.operators.make_operator(_JoinImpl, _combined_schemas)

lale.docstrings.set_docstrings(Join)



from numpy import inf, nan
from sklearn.linear_model import ElasticNetCV as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _ElasticNetCVImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for ElasticNetCV    Elastic Net model with iterative fitting along a regularization path.",
    "allOf": [
        {
            "type": "object",
            "required": [
                "l1_ratio",
                "eps",
                "n_alphas",
                "alphas",
                "fit_intercept",
                "normalize",
                "precompute",
                "max_iter",
                "tol",
                "cv",
                "copy_X",
                "verbose",
                "n_jobs",
                "positive",
                "random_state",
                "selection",
            ],
            "relevantToOptimizer": [
                "l1_ratio",
                "eps",
                "n_alphas",
                "alpha",
                "fit_intercept",
                "normalize",
                "precompute",
                "max_iter",
                "tol",
                "cv",
                "copy_X",
                "positive",
                "selection",
            ],
            "additionalProperties": False,
            "properties": {
                "l1_ratio": {
                    "XXX TODO XXX": "float or array of floats, optional",
                    "description": "float between 0 and 1 passed to ElasticNet (scaling between l1 and l2 penalties)",
                    "type": "number",
                    "minimumForOptimizer": 0.0,
                    "maximumForOptimizer": 1.0,
                    "distribution": "uniform",
                    "default": 0.5,
                },
                "eps": {
                    "type": "number",
                    "minimumForOptimizer": 0.001,
                    "maximumForOptimizer": 0.1,
                    "distribution": "loguniform",
                    "default": 0.001,
                    "description": "Length of the path",
                },
                "n_alphas": {
                    "type": "integer",
                    "minimumForOptimizer": 100,
                    "maximumForOptimizer": 101,
                    "distribution": "uniform",
                    "default": 100,
                    "description": "Number of alphas along the regularization path, used for each l1_ratio.",
                },
                "alphas": {
                    "anyOf": [
                        {
                            "type": "array",
                            "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                            "XXX TODO XXX": "numpy array, optional",
                        },
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "List of alphas where to compute the models",
                },
                "fit_intercept": {
                    "type": "boolean",
                    "default": True,
                    "description": "whether to calculate the intercept for this model",
                },
                "normalize": {
                    "type": "boolean",
                    "default": False,
                    "description": "This parameter is ignored when ``fit_intercept`` is set to False",
                },
                "precompute": {
                    "anyOf": [
                        {
                            "type": "array",
                            "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                            "XXX TODO XXX": "True | False | 'auto' | array-like",
                            "forOptimizer": False,
                        },
                        {"enum": ["auto"]},
                    ],
                    "default": "auto",
                    "description": "Whether to use a precomputed Gram matrix to speed up calculations",
                },
                "max_iter": {
                    "type": "integer",
                    "minimumForOptimizer": 10,
                    "maximumForOptimizer": 1000,
                    "distribution": "uniform",
                    "default": 1000,
                    "description": "The maximum number of iterations",
                },
                "tol": {
                    "type": "number",
                    "minimumForOptimizer": 1e-08,
                    "maximumForOptimizer": 0.01,
                    "distribution": "loguniform",
                    "default": 0.0001,
                    "description": "The tolerance for the optimization: if the updates are smaller than ``tol``, the optimization code checks the dual gap for optimality and continues until it is smaller than ``tol``.",
                },
                "cv": {
                    "description": """Cross-validation as integer or as object that has a split function.
                        The fit method performs cross validation on the input dataset for per
                        trial, and uses the mean cross validation performance for optimization.
                        This behavior is also impacted by handle_cv_failure flag.
                        If integer: number of folds in sklearn.model_selection.StratifiedKFold.
                        If object with split function: generator yielding (train, test) splits
                        as arrays of indices. Can use any of the iterators from
                        https://scikit-learn.org/stable/modules/cross_validation.html#cross-validation-iterators.""",
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimum": 1,
                            "default": 5,
                            "minimumForOptimizer": 3,
                            "maximumForOptimizer": 4,
                            "distribution": "uniform",
                        },
                        {"laleType": "Any", "forOptimizer": False},
                    ],
                },
                "copy_X": {
                    "type": "boolean",
                    "default": True,
                    "description": "If ``True``, X will be copied; else, it may be overwritten.",
                },
                "verbose": {
                    "anyOf": [{"type": "boolean"}, {"type": "integer"}],
                    "default": 0,
                    "description": "Amount of verbosity.",
                },
                "n_jobs": {
                    "anyOf": [{"type": "integer"}, {"enum": [None]}],
                    "default": 1,
                    "description": "Number of CPUs to use during the cross validation",
                },
                "positive": {
                    "type": "boolean",
                    "default": False,
                    "description": "When set to ``True``, forces the coefficients to be positive.",
                },
                "random_state": {
                    "anyOf": [
                        {"type": "integer"},
                        {"laleType": "numpy.random.RandomState"},
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "The seed of the pseudo random number generator that selects a random feature to update",
                },
                "selection": {
                    "enum": ["random", "cyclic"],
                    "default": "cyclic",
                    "description": "If set to 'random', a random coefficient is updated every iteration rather than looping over features sequentially by default",
                },
            },
        },
        {
            "description": "From /linear_model/_coordinate_descent.py:None:_alpha_grid, Exception: raise ValueError(     "
            "Automatic alpha grid generation is not supported for l1_ratio=0. Please supply a grid by providing your estimator with the appropriate `alphas=` argument.",
            "anyOf": [
                {"type": "object", "properties": {"alphas": {"not": {"enum": [None]}}}},
                {"type": "object", "properties": {"l1_ratio": {"not": {"enum": [0]}}}},
            ],
        },
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Fit linear model with coordinate descent",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "laleType": "Any",
            "XXX TODO XXX": "{array-like}, shape (n_samples, n_features)",
            "description": "Training data",
        },
        "y": {
            "anyOf": [
                {"type": "array", "items": {"type": "number"}},
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "Target values",
        },
    },
}
_input_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predict using the linear model",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                    "XXX TODO XXX": "array_like or sparse matrix, shape (n_samples, n_features)",
                },
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "Samples.",
        }
    },
}
_output_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Returns predicted values.",
    "type": "array",
    "items": {"type": "number"},
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.linear_model.ElasticNetCV#sklearn-linear_model-elasticnetcv",
    "import_from": "sklearn.linear_model",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "regressor"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}
ElasticNetCV = make_operator(_ElasticNetCVImpl, _combined_schemas)

set_docstrings(ElasticNetCV)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import warnings
from test import EnableSchemaValidation

import jsonschema
from sklearn.datasets import load_iris

import lale.lib.lale
import lale.type_checking
from lale.lib.lale import NoOp
from lale.lib.sklearn import (
    PCA,
    SVC,
    IsolationForest,
    KMeans,
    KNeighborsClassifier,
    LogisticRegression,
    MLPClassifier,
    Nystroem,
    PassiveAggressiveClassifier,
    RidgeClassifier,
    SGDClassifier,
    SimpleImputer,
    VotingClassifier,
)
from lale.search.lale_grid_search_cv import get_grid_search_parameter_grids
from lale.settings import set_disable_data_schema_validation

set_disable_data_schema_validation(False)


class TestClassification(unittest.TestCase):
    def setUp(self):
        from sklearn.model_selection import train_test_split

        data = load_iris()
        X, y = data.data, data.target
        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(X, y)


def create_function_test_classifier(clf_name):
    def test_classifier(self):
        X_train, y_train = self.X_train, self.y_train
        import importlib

        module_name = ".".join(clf_name.split(".")[0:-1])
        class_name = clf_name.split(".")[-1]
        module = importlib.import_module(module_name)

        class_ = getattr(module, class_name)
        clf = class_()

        # test_schemas_are_schemas
        lale.type_checking.validate_is_schema(clf.input_schema_fit())
        lale.type_checking.validate_is_schema(clf.input_schema_predict())
        lale.type_checking.validate_is_schema(clf.output_schema_predict())
        lale.type_checking.validate_is_schema(clf.hyperparam_schema())

        # test_init_fit_predict
        trained = clf.fit(self.X_train, self.y_train)
        _ = trained.predict(self.X_test)

        # test score
        if not isinstance(
            clf, IsolationForest  # type: ignore
        ):  # IsolationForest does not define score
            _ = trained.score(self.X_test, self.y_test)

        from lale.lib.sklearn.gradient_boosting_classifier import (
            GradientBoostingClassifier,
        )

        if isinstance(clf, GradientBoostingClassifier):  # type: ignore
            # because exponential loss does not work with iris dataset as it is not binary classification
            import lale.schemas as schemas

            clf = clf.customize_schema(
                loss=schemas.Enum(default="deviance", values=["deviance"])
            )

        # test_with_hyperopt
        from lale.lib.lale import Hyperopt

        hyperopt = Hyperopt(estimator=clf, max_evals=1, verbose=True)
        trained = hyperopt.fit(self.X_train, self.y_train)
        _ = trained.predict(self.X_test)
        # test_cross_validation
        from lale.helpers import cross_val_score

        cv_results = cross_val_score(clf, X_train, y_train, cv=2)
        self.assertEqual(len(cv_results), 2)

        # test_with_gridsearchcv_auto_wrapped
        from sklearn.metrics import accuracy_score, make_scorer

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            grid_search = lale.lib.lale.GridSearchCV(
                estimator=clf,
                lale_num_samples=1,
                lale_num_grids=1,
                cv=2,
                scoring=make_scorer(accuracy_score),
            )
            grid_search.fit(X_train, y_train)

        # test_predict_on_trainable
        trained = clf.fit(X_train, y_train)
        clf.predict(X_train)

        # test_to_json
        clf.to_json()

        # test_in_a_pipeline
        pipeline = NoOp() >> clf
        trained = pipeline.fit(self.X_train, self.y_train)
        _ = trained.predict(self.X_test)

    test_classifier.__name__ = "test_{0}".format(clf.split(".")[-1])
    return test_classifier


classifiers = [
    "lale.lib.sklearn.DummyClassifier",
    "lale.lib.sklearn.RandomForestClassifier",
    "lale.lib.sklearn.DecisionTreeClassifier",
    "lale.lib.sklearn.ExtraTreesClassifier",
    "lale.lib.sklearn.GradientBoostingClassifier",
    "lale.lib.sklearn.GaussianNB",
    "lale.lib.sklearn.QuadraticDiscriminantAnalysis",
    "lale.lib.lightgbm.LGBMClassifier",
    "lale.lib.xgboost.XGBClassifier",
    "lale.lib.sklearn.KNeighborsClassifier",
    "lale.lib.sklearn.LinearSVC",
    "lale.lib.sklearn.LogisticRegression",
    "lale.lib.sklearn.MLPClassifier",
    "lale.lib.sklearn.SVC",
    "lale.lib.sklearn.PassiveAggressiveClassifier",
    "lale.lib.sklearn.MultinomialNB",
    "lale.lib.sklearn.AdaBoostClassifier",
    "lale.lib.sklearn.SGDClassifier",
    "lale.lib.sklearn.RidgeClassifier",
    "lale.lib.sklearn.IsolationForest",
    "lale.lib.sklearn.KMeans",
]
for clf in classifiers:
    setattr(
        TestClassification,
        "test_{0}".format(clf.split(".")[-1]),
        create_function_test_classifier(clf),
    )


class TestMLPClassifier(unittest.TestCase):
    def test_with_multioutput_targets(self):
        import numpy as np
        from sklearn.datasets import make_classification
        from sklearn.utils import shuffle

        X, y1 = make_classification(
            n_samples=10, n_features=100, n_informative=30, n_classes=3, random_state=1
        )
        y2 = shuffle(y1, random_state=1)
        y3 = shuffle(y1, random_state=2)
        Y = np.vstack((y1, y2, y3)).T
        trainable = KNeighborsClassifier()
        trained = trainable.fit(X, Y)
        _ = trained.predict(X)

    def test_predict_proba(self):
        trainable = MLPClassifier()
        iris = load_iris()
        trained = trainable.fit(iris.data, iris.target)
        #        with self.assertWarns(DeprecationWarning):
        _ = trainable.predict_proba(iris.data)
        _ = trained.predict_proba(iris.data)


class TestVotingClassifier(unittest.TestCase):
    def setUp(self):
        from sklearn.model_selection import train_test_split

        data = load_iris()
        X, y = data.data, data.target
        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(X, y)
        import warnings

        warnings.filterwarnings("ignore")

    def test_with_lale_classifiers(self):
        clf = VotingClassifier(
            estimators=[("knn", KNeighborsClassifier()), ("lr", LogisticRegression())]
        )
        trained = clf.fit(self.X_train, self.y_train)
        trained.predict(self.X_test)

    def test_with_lale_pipeline(self):
        from lale.lib.sklearn import VotingClassifier

        clf = VotingClassifier(
            estimators=[
                ("knn", KNeighborsClassifier()),
                ("pca_lr", PCA() >> LogisticRegression()),
            ]
        )
        trained = clf.fit(self.X_train, self.y_train)
        trained.predict(self.X_test)

    def test_with_hyperopt(self):
        from lale.lib.lale import Hyperopt
        from lale.lib.sklearn import VotingClassifier

        clf = VotingClassifier(
            estimators=[("knn", KNeighborsClassifier()), ("lr", LogisticRegression())]
        )
        _ = clf.auto_configure(self.X_train, self.y_train, Hyperopt, max_evals=1)

    def test_with_gridsearch(self):
        from sklearn.metrics import accuracy_score, make_scorer

        from lale.lib.lale import GridSearchCV
        from lale.lib.sklearn import VotingClassifier

        clf = VotingClassifier(
            estimators=[("knn", KNeighborsClassifier()), ("rc", RidgeClassifier())],
            voting="hard",
        )
        _ = clf.auto_configure(
            self.X_train,
            self.y_train,
            GridSearchCV,
            lale_num_samples=1,
            lale_num_grids=1,
            cv=2,
            scoring=make_scorer(accuracy_score),
        )

    @unittest.skip("TODO: get this working with sklearn 0.23")
    def test_with_observed_gridsearch(self):
        from sklearn.metrics import accuracy_score, make_scorer

        from lale.lib.lale import GridSearchCV
        from lale.lib.lale.observing import LoggingObserver
        from lale.lib.sklearn import VotingClassifier

        clf = VotingClassifier(
            estimators=[("knn", KNeighborsClassifier()), ("rc", RidgeClassifier())],
            voting="hard",
        )
        _ = clf.auto_configure(
            self.X_train,
            self.y_train,
            GridSearchCV,
            lale_num_samples=1,
            lale_num_grids=1,
            cv=2,
            scoring=make_scorer(accuracy_score),
            observer=LoggingObserver,
        )


class TestBaggingClassifier(unittest.TestCase):
    def setUp(self):
        from sklearn.model_selection import train_test_split

        data = load_iris()
        X, y = data.data, data.target
        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(X, y)

        import warnings

        warnings.filterwarnings("ignore")

    def test_with_lale_classifiers(self):
        from lale.lib.sklearn import BaggingClassifier

        clf = BaggingClassifier(base_estimator=LogisticRegression())
        trained = clf.fit(self.X_train, self.y_train)
        trained.predict(self.X_test)

    def test_with_lale_pipeline(self):
        from lale.lib.sklearn import BaggingClassifier

        clf = BaggingClassifier(base_estimator=PCA() >> LogisticRegression())
        trained = clf.fit(self.X_train, self.y_train)
        trained.predict(self.X_test)

    def test_with_hyperopt(self):
        from lale.lib.lale import Hyperopt
        from lale.lib.sklearn import BaggingClassifier

        clf = BaggingClassifier(base_estimator=LogisticRegression())
        trained = clf.auto_configure(self.X_train, self.y_train, Hyperopt, max_evals=1)
        print(trained.to_json())

    def test_pipeline_with_hyperopt(self):
        from lale.lib.lale import Hyperopt
        from lale.lib.sklearn import BaggingClassifier

        clf = BaggingClassifier(base_estimator=PCA() >> LogisticRegression())
        _ = clf.auto_configure(self.X_train, self.y_train, Hyperopt, max_evals=1)

    def test_pipeline_choice_with_hyperopt(self):
        from lale.lib.lale import Hyperopt
        from lale.lib.sklearn import BaggingClassifier

        clf = BaggingClassifier(
            base_estimator=PCA() >> (LogisticRegression() | KNeighborsClassifier())
        )
        _ = clf.auto_configure(self.X_train, self.y_train, Hyperopt, max_evals=1)

    def test_predict_log_proba(self):
        from lale.lib.sklearn import BaggingClassifier

        clf = BaggingClassifier(base_estimator=PCA() >> LogisticRegression())
        trained = clf.fit(self.X_train, self.y_train)
        trained.predict_log_proba(self.X_test)

    def test_predict_log_proba_trained_trainable(self):
        from lale.lib.sklearn import BaggingClassifier

        clf = BaggingClassifier()
        clf.fit(self.X_train, self.y_train)
        with self.assertWarns(DeprecationWarning):
            clf.predict_log_proba(self.X_test)

    def test_predict_log_proba_trainable(self):
        from lale.lib.sklearn import BaggingClassifier

        clf = BaggingClassifier(base_estimator=PCA() >> LogisticRegression())
        with self.assertRaises(ValueError):
            clf.predict_log_proba(self.X_test)


class TestStackingClassifier(unittest.TestCase):
    def setUp(self):
        from sklearn.model_selection import train_test_split

        data = load_iris()
        X, y = data.data, data.target
        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(X, y)

    def test_with_lale_classifiers(self):
        from lale.lib.sklearn import StackingClassifier

        clf = StackingClassifier(estimators=[("base", LogisticRegression())])
        trained = clf.fit(self.X_train, self.y_train)
        trained.predict(self.X_test)

    def test_with_lale_pipeline(self):
        from lale.lib.sklearn import StackingClassifier

        clf = StackingClassifier(estimators=[("base", PCA() >> LogisticRegression())])
        trained = clf.fit(self.X_train, self.y_train)
        trained.predict(self.X_test)

    def test_with_hyperopt(self):
        from lale.lib.lale import Hyperopt
        from lale.lib.sklearn import StackingClassifier

        clf = StackingClassifier(estimators=[("base", LogisticRegression())])
        trained = clf.auto_configure(self.X_train, self.y_train, Hyperopt, max_evals=1)
        print(trained.to_json())

    def test_pipeline_with_hyperopt(self):
        from lale.lib.lale import Hyperopt
        from lale.lib.sklearn import StackingClassifier

        clf = StackingClassifier(estimators=[("base", PCA() >> LogisticRegression())])
        _ = clf.auto_configure(self.X_train, self.y_train, Hyperopt, max_evals=1)

    def test_pipeline_choice_with_hyperopt(self):
        from lale.lib.lale import Hyperopt
        from lale.lib.sklearn import StackingClassifier

        clf = StackingClassifier(
            estimators=[
                ("base", PCA() >> (LogisticRegression() | KNeighborsClassifier()))
            ]
        )
        _ = clf.auto_configure(self.X_train, self.y_train, Hyperopt, max_evals=1)


class TestSpuriousSideConstraintsClassification(unittest.TestCase):
    # This was prompted buy a bug, keeping it as it may help with support for other sklearn versions
    def setUp(self):
        from sklearn.model_selection import train_test_split

        data = load_iris()
        X, y = data.data, data.target
        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(X, y)

    def test_sgd_classifier(self):
        reg = SGDClassifier(loss="squared_loss", epsilon=0.2)
        reg.fit(self.X_train, self.y_train)

    def test_sgd_classifier_1(self):
        reg = SGDClassifier(learning_rate="optimal", eta0=0.2)
        reg.fit(self.X_train, self.y_train)

    def test_sgd_classifier_2(self):
        reg = SGDClassifier(early_stopping=False, validation_fraction=0.2)
        reg.fit(self.X_train, self.y_train)

    def test_sgd_classifier_3(self):
        reg = SGDClassifier(l1_ratio=0.2, penalty="l1")
        reg.fit(self.X_train, self.y_train)

    def test_mlp_classifier(self):
        reg = MLPClassifier(early_stopping=False, validation_fraction=0.2)
        reg.fit(self.X_train, self.y_train)

    def test_mlp_classifier_1(self):
        reg = MLPClassifier(beta_1=0.8, solver="sgd")
        reg.fit(self.X_train, self.y_train)

    def test_mlp_classifier_2b(self):
        reg = MLPClassifier(beta_2=0.8, solver="sgd")
        reg.fit(self.X_train, self.y_train)

    def test_mlp_classifier_2e(self):
        reg = MLPClassifier(epsilon=0.8, solver="sgd")
        reg.fit(self.X_train, self.y_train)

    def test_mlp_classifier_3(self):
        reg = MLPClassifier(n_iter_no_change=100, solver="lbfgs")
        reg.fit(self.X_train, self.y_train)

    def test_mlp_classifier_4(self):
        reg = MLPClassifier(early_stopping=True, solver="lbfgs")
        reg.fit(self.X_train, self.y_train)

    def test_mlp_classifier_5(self):
        reg = MLPClassifier(nesterovs_momentum=False, solver="lbfgs")
        reg.fit(self.X_train, self.y_train)

    def test_mlp_classifier_6(self):
        reg = MLPClassifier(momentum=0.8, solver="lbfgs")
        reg.fit(self.X_train, self.y_train)

    def test_mlp_classifier_7(self):
        reg = MLPClassifier(shuffle=False, solver="lbfgs")
        reg.fit(self.X_train, self.y_train)

    def test_mlp_classifier_8(self):
        reg = MLPClassifier(learning_rate="invscaling", solver="lbfgs")
        reg.fit(self.X_train, self.y_train)

    def test_mlp_classifier_9(self):
        reg = MLPClassifier(learning_rate_init=0.002, solver="lbfgs")
        reg.fit(self.X_train, self.y_train)

    def test_mlp_classifier_10(self):
        reg = MLPClassifier(learning_rate="invscaling", power_t=0.4, solver="lbfgs")
        reg.fit(self.X_train, self.y_train)

    def test_passive_aggressive_classifier(self):
        reg = PassiveAggressiveClassifier(validation_fraction=0.4, early_stopping=False)
        reg.fit(self.X_train, self.y_train)

    def test_svc(self):
        reg = SVC(kernel="linear", gamma=1)
        reg.fit(self.X_train, self.y_train)

    def test_simple_imputer(self):
        reg = SimpleImputer(strategy="mean", fill_value=10)
        reg.fit(self.X_train, self.y_train)

    def test_nystroem(self):
        reg = Nystroem(kernel="cosine", gamma=0.1)
        reg.fit(self.X_train, self.y_train)

    def test_nystroem_1(self):
        reg = Nystroem(kernel="cosine", coef0=0.1)
        reg.fit(self.X_train, self.y_train)

    def test_nystroem_2(self):
        reg = Nystroem(kernel="cosine", degree=2)
        reg.fit(self.X_train, self.y_train)

    def test_ridge_classifier(self):
        reg = RidgeClassifier(fit_intercept=False)
        reg.fit(self.X_train, self.y_train)

    def test_ridge_classifier_1(self):
        reg = RidgeClassifier(solver="svd", max_iter=10)
        reg.fit(self.X_train, self.y_train)


class TestKNeighborsClassifier(unittest.TestCase):
    def test_with_multioutput_targets(self):
        import numpy as np
        from sklearn.datasets import make_classification
        from sklearn.utils import shuffle

        X, y1 = make_classification(
            n_samples=10, n_features=100, n_informative=30, n_classes=3, random_state=1
        )
        y2 = shuffle(y1, random_state=1)
        y3 = shuffle(y1, random_state=2)
        Y = np.vstack((y1, y2, y3)).T
        trainable = KNeighborsClassifier()
        trained = trainable.fit(X, Y)
        _ = trained.predict(X)

    def test_predict_proba(self):
        trainable = KNeighborsClassifier()
        iris = load_iris()
        trained = trainable.fit(iris.data, iris.target)
        # with self.assertWarns(DeprecationWarning):
        _ = trainable.predict_proba(iris.data)
        _ = trained.predict_proba(iris.data)


class TestLogisticRegression(unittest.TestCase):
    def test_hyperparam_keyword_enum(self):
        _ = LogisticRegression(
            LogisticRegression.enum.penalty.l1,
            C=0.1,
            solver=LogisticRegression.enum.solver.saga,
        )

    def test_hyperparam_exclusive_min(self):
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                _ = LogisticRegression(LogisticRegression.enum.penalty.l1, C=0.0)

    def test_hyperparam_penalty_solver_dependence(self):
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                _ = LogisticRegression(
                    LogisticRegression.enum.penalty.l1,
                    LogisticRegression.enum.solver.newton_cg,
                )

    def test_hyperparam_dual_penalty_solver_dependence(self):
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                _ = LogisticRegression(
                    LogisticRegression.enum.penalty.l2,
                    LogisticRegression.enum.solver.sag,
                    dual=True,
                )

    def test_sample_weight(self):
        import numpy as np

        trainable_lr = LogisticRegression(n_jobs=1)
        iris = load_iris()
        trained_lr = trainable_lr.fit(
            iris.data, iris.target, sample_weight=np.arange(len(iris.target))
        )
        _ = trained_lr.predict(iris.data)

    def test_predict_proba(self):
        import numpy as np

        trainable_lr = LogisticRegression(n_jobs=1)
        iris = load_iris()
        trained_lr = trainable_lr.fit(
            iris.data, iris.target, sample_weight=np.arange(len(iris.target))
        )
        # with self.assertWarns(DeprecationWarning):
        _ = trainable_lr.predict_proba(iris.data)
        _ = trained_lr.predict_proba(iris.data)

    def test_decision_function(self):
        import numpy as np

        trainable_lr = LogisticRegression(n_jobs=1)
        iris = load_iris()
        trained_lr = trainable_lr.fit(
            iris.data, iris.target, sample_weight=np.arange(len(iris.target))
        )
        _ = trained_lr.decision_function(iris.data)

    def test_with_sklearn_gridsearchcv(self):
        from sklearn.datasets import load_iris
        from sklearn.metrics import accuracy_score, make_scorer
        from sklearn.model_selection import GridSearchCV

        lr = LogisticRegression()
        parameters = {"solver": ("liblinear", "lbfgs"), "penalty": ["l2"]}
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            clf = GridSearchCV(
                lr, parameters, cv=5, scoring=make_scorer(accuracy_score)
            )
            iris = load_iris()
            clf.fit(iris.data, iris.target)

    def test_with_randomizedsearchcv(self):
        import numpy as np
        from scipy.stats.distributions import uniform
        from sklearn.metrics import accuracy_score, make_scorer
        from sklearn.model_selection import RandomizedSearchCV

        lr = LogisticRegression()
        ranges, cat_idx = lr.get_param_ranges()
        # specify parameters and distributions to sample from
        # the loguniform distribution needs to be taken care of properly
        param_dist = {"solver": ranges["solver"], "C": uniform(0.03125, np.log(32768))}
        # run randomized search
        n_iter_search = 5
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            random_search = RandomizedSearchCV(
                lr,
                param_distributions=param_dist,
                n_iter=n_iter_search,
                cv=5,
                scoring=make_scorer(accuracy_score),
            )
            iris = load_iris()
            random_search.fit(iris.data, iris.target)

    def test_grid_search_on_trained(self):
        from sklearn.metrics import accuracy_score, make_scorer
        from sklearn.model_selection import GridSearchCV

        iris = load_iris()
        X, y = iris.data, iris.target
        lr = LogisticRegression()
        trained = lr.fit(X, y)
        parameters = {"solver": ("liblinear", "lbfgs"), "penalty": ["l2"]}

        _ = GridSearchCV(trained, parameters, cv=5, scoring=make_scorer(accuracy_score))

    def test_grid_search_on_trained_auto(self):
        from sklearn.metrics import accuracy_score, make_scorer
        from sklearn.model_selection import GridSearchCV

        iris = load_iris()
        X, y = iris.data, iris.target
        lr = LogisticRegression()
        trained = lr.fit(X, y)
        parameters = get_grid_search_parameter_grids(lr, num_samples=2)

        _ = GridSearchCV(trained, parameters, cv=5, scoring=make_scorer(accuracy_score))

    def test_doc(self):
        from test.mock_custom_operators import MyLR

        import sklearn.datasets
        import sklearn.utils

        iris = load_iris()
        X_all, y_all = sklearn.utils.shuffle(iris.data, iris.target, random_state=42)
        X_train, y_train = X_all[10:], y_all[10:]
        X_test, y_test = X_all[:10], y_all[:10]
        print("expected {}".format(y_test))
        import warnings

        warnings.filterwarnings("ignore", category=FutureWarning)
        trainable = MyLR(solver="lbfgs", C=0.1)
        trained = trainable.fit(X_train, y_train)
        predictions = trained.predict(X_test)
        print("actual {}".format(predictions))


class TestIsolationForest(unittest.TestCase):
    def setUp(self):
        from sklearn.datasets import load_boston
        from sklearn.model_selection import train_test_split

        data = load_boston()
        X, y = data.data, data.target
        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(X, y)
        import warnings

        warnings.filterwarnings("ignore")

    def test_with_no_y(self):
        clf = IsolationForest()
        trained = clf.fit(self.X_train)
        trained.predict(self.X_test)

    def test_with_hyperopt(self):
        def my_scorer(estimator, X, y=None):
            return 1

        from lale.lib.lale import Hyperopt

        hyperopt = Hyperopt(
            estimator=IsolationForest(max_features=1.0, max_samples=1.0),
            max_evals=5,
            verbose=True,
            scoring=my_scorer,
        )
        trained = hyperopt.fit(self.X_train)
        _ = trained.predict(self.X_test)

    def test_decision_function_1(self):
        def my_scorer(estimator, X, y=None):
            return 1

        from lale.lib.lale import Hyperopt

        hyperopt = Hyperopt(
            estimator=IsolationForest(max_features=1.0, max_samples=1.0),
            max_evals=5,
            verbose=True,
            scoring=my_scorer,
        )
        trained = hyperopt.fit(self.X_train)
        pipeline = trained.get_pipeline()
        assert pipeline is not None
        _ = pipeline.decision_function(self.X_test)

    def test_decision_function_2(self):
        def my_scorer(estimator, X, y=None):
            return 1

        from lale.lib.lale import Hyperopt
        from lale.lib.sklearn import MinMaxScaler

        hyperopt = Hyperopt(
            estimator=MinMaxScaler()
            >> IsolationForest(max_features=1.0, max_samples=1.0),
            max_evals=5,
            verbose=True,
            scoring=my_scorer,
        )
        trained = hyperopt.fit(self.X_train)
        pipeline = trained.get_pipeline()
        assert pipeline is not None
        _ = pipeline.decision_function(self.X_test)

    def test_score_samples(self):
        clf = IsolationForest()
        trained = clf.fit(self.X_train)
        trained.score_samples(self.X_test)

    def test_score_samples_trainable(self):
        clf = IsolationForest()
        with self.assertRaises(ValueError):
            clf.score_samples(self.X_test)

    def test_score_samples_trained_trainable(self):
        clf = IsolationForest()
        clf.fit(self.X_train)
        with self.assertWarns(DeprecationWarning):
            clf.score_samples(self.X_test)


class TestKMeans(unittest.TestCase):
    def setUp(self):
        from sklearn.datasets import load_boston
        from sklearn.model_selection import train_test_split

        data = load_boston()
        X, y = data.data, data.target
        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(X, y)
        import warnings

        warnings.filterwarnings("ignore")

    def test_with_no_y(self):
        clf = KMeans()
        trained = clf.fit(self.X_train)
        trained.predict(self.X_test)

    def test_with_hyperopt(self):
        from lale.lib.lale import Hyperopt

        def my_scorer(estimator, X, y=None):
            return 1

        hyperopt = Hyperopt(
            estimator=KMeans(n_clusters=3), max_evals=5, verbose=True, scoring=my_scorer
        )
        trained = hyperopt.fit(self.X_train)
        _ = trained.predict(self.X_test)
